/* GKrellM
|  Copyright (C) 1999-2001 Bill Wilson
|
|  Author:  Bill Wilson    bill@gkrellm.net
|  Latest versions might be found at:  http://gkrellm.net
|
|  This program is free software which I release under the GNU General Public
|  License. You may redistribute and/or modify this program under the terms
|  of that license as published by the Free Software Foundation; either
|  version 2 of the License, or (at your option) any later version.
|
|  This program is distributed in the hope that it will be useful,
|  but WITHOUT ANY WARRANTY; without even the implied warranty of
|  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
|  GNU General Public License for more details.
| 
|  To get a copy of the GNU General Puplic License, write to the Free Software
|  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* 10/12/2000	NetBSD code contributed by Anthony Mallet
|			<metall@ficus.yi.org>
| 2/25/2000	FreeBSD code contributed by Hajimu UMEMOTO ume@mahoroba.org
*/

#include "gkrellm.h"
#include "gkrellm_private_proto.h"


typedef struct
	{
	gchar	*directory;
	gchar	*device;
	gchar	*type;
	gchar	*options;
	}
	Mount;

typedef struct
	{
	GtkWidget		*vbox;
	gint			idx;
	Panel			panel;
	DecalButton		*md_button;
	Decal			*mount_decal,
					*text_decal;
	gchar			*label;
	gint			label_is_data;
	gint			restore_label;

	Mount			mount;
	gint			enable_mounting;

	Launcher		launch_mount,
					launch_umount;

	/* Monitors in the drawer will not be visible if the drawer is shut
	|  unless the show_if_mounted option is set and the monitor is
	|  mounted.
	*/
	gint			in_drawer;
	gint			show_if_mounted;	/* for in_drawer monitors */
	gint			is_visible;

	gint			is_mounted;
	gint			is_nfs_fs;
	gint			eject_pending;
	GString			*pipe_gstring;		/* output of mount commands */

	gulong			blocks,
					bfree,
					bavail,
					bsize;
	}
	FSmon;


static GList		*fs_mon_list,
					*proc_mounts_list;
static GList		*fstab_list;

static gint			uid;

void	(*get_mountlist)(),
		(*get_fsusage)(),
		(*get_fstab_list)();

/* ====== System dependent interface ====================================== */


/* ----- FreeBSD & NetBSD ----------------------------------------------- */

#if defined(__FreeBSD__) || defined(__NetBSD__) || defined(__OpenBSD__)
#include <sys/mount.h>
#include <sys/cdio.h>

#if defined(__FreeBSD__)
# include <osreldate.h>
#if __FreeBSD_version < 300000
static char	*mnttype[] = INITMOUNTNAMES;
#endif
#endif

static void
get_bsd_mounts_list()
	{
	Mount		*m;
	gchar		*s, *dev, *dir, *type;
	struct statfs	*mntbuf;
	gint		mntsize, i;

	while (proc_mounts_list)
		{
		m = (Mount *) proc_mounts_list->data;
		g_free(m->directory);
		g_free(m->device);
		g_free(m->type);
		g_free(proc_mounts_list->data);
		proc_mounts_list = g_list_remove(proc_mounts_list,
						 proc_mounts_list->data);
		}
	if ((mntsize = getmntinfo(&mntbuf, MNT_NOWAIT)) == 0)
		return;
	for (i = 0; i < mntsize; i++)
		{
#if defined(__FreeBSD__) && __FreeBSD_version < 300000
		type = mnttype[mntbuf[i].f_type];
#else
		type = mntbuf[i].f_fstypename;
#endif
		dir = mntbuf[i].f_mntonname;
		dev = mntbuf[i].f_mntfromname;
		/* Strip trailing / from the directory.
		*/
		s = strrchr(dir, (int) '/');
		if (s && s != dir && *(s+1) == '\0')
			*s = '\0';
		m = g_new0(Mount, 1);
		m->directory = g_strdup(dir);
		m->device = g_strdup(dev);
		m->type = g_strdup(type);

		proc_mounts_list = g_list_append(proc_mounts_list, m);
		}
	}
#endif


/* ----- Linux ------------------------------------------------------ */

#if defined(__linux__)
#include <sys/vfs.h>
#include <linux/cdrom.h>
#include <mntent.h>

#define	PROC_MOUNTS_FILE	"/proc/mounts"

  /* A list of mounted file systems can be read from /proc/mounts or
  |  /etc/mtab (getmntent).  Using /proc/mounts eliminates disk accesses,
  |  but for some reason /proc/mounts reports a "." for the mounted
  |  directory for smbfs types.  So I use /proc/mounts with a fallback
  |  to using getmntent().
  */
#if !defined (_PATH_MOUNTED)
#define _PATH_MOUNTED   "/etc/mtab"
#endif

static void
getmntent_fallback(Mount *m)
	{
	FILE			*f;
	struct mntent	*mnt;

	if ((f = setmntent(_PATH_MOUNTED, "r")) == NULL)
		return;
	while ((mnt = getmntent(f)) != NULL)
		{
		if (!strcmp(m->device, mnt->mnt_fsname))
			{
			gkrellm_dup_string(&m->directory, mnt->mnt_dir);
			break;
			}
		}
	endmntent(f);
	}

static void
get_linux_proc_mounts_list()
	{
	FILE		*f;
	Mount		*m;
	gchar		*s, buf[128], dev[64], dir[128], type[32];

	while (proc_mounts_list)
		{
		m = (Mount *) proc_mounts_list->data;
		g_free(m->directory);
		g_free(m->device);
		g_free(m->type);
		g_free(proc_mounts_list->data);
		proc_mounts_list =
				g_list_remove(proc_mounts_list, proc_mounts_list->data);
		}
	if ((f = fopen(PROC_MOUNTS_FILE, "r")) == NULL)
		return;
	while (fgets(buf, sizeof(buf), f))
		{
		sscanf(buf, "%63s %127s %31s", dev, dir, type);
		if (   strcmp(type, "devpts") == 0
			|| strcmp(type, "proc") == 0
		   )
			continue;
		/* Strip trailing / from the directory.
		*/
		s = strrchr(dir, (int) '/');
		if (s && s != dir && *(s+1) == '\0')
			*s = '\0';
		m = g_new0(Mount, 1);
		m->directory = g_strdup(dir);
		m->device = g_strdup(dev);
		m->type = g_strdup(type);
		if (dir[0] == '.')
			getmntent_fallback(m);

		proc_mounts_list = g_list_append(proc_mounts_list, m);
		}
	fclose(f);
	}
#endif	/* __linux__ */


/* ----- FreeBSD, NetBSD and Linux -------------------------------------- */

#if defined(__linux__) || defined(__FreeBSD__) || defined(__NetBSD__) \
	|| defined(__OpenBSD__)

static void
get_statfs_fsusage(FSmon *fs)
	{
	struct statfs	st;

	if (statfs(fs->mount.directory, &st) == 0)
		{
		fs->blocks = (gulong) st.f_blocks;
		fs->bavail = (gulong) st.f_bavail;
		fs->bfree =  (gulong) st.f_bfree;
		fs->bsize  = (gulong) st.f_bsize;
		}
	else
		{
		fs->blocks = 0;
		fs->bavail = 0;
		fs->bfree =  0;
		fs->bsize  = 0;
		}
	}
#endif


/* ----- Generic /etc/fstab -------------------------------------------- */
static gchar *
fix_fstab_name(gchar *buf)
	{
	gchar	*rp,
			*wp;

	if (buf == NULL)
		return g_strdup("");
	rp = buf;
	wp = buf;
	do	/* This loop same as in libc6 getmntent()	*/
		if (rp[0] == '\\' && rp[1] == '0' && rp[2] == '4' && rp[3] == '0')
			{
			*wp++ = ' ';		/* \040 is a SPACE.  */
			rp += 3;
			}
		else if (rp[0] == '\\' && rp[1] == '0' && rp[2] == '1' && rp[3] == '2')
			{
			*wp++ = '\t';		/* \012 is a TAB.  */
			rp += 3;
			}
		else if (rp[0] == '\\' && rp[1] == '\\')
			{
			*wp++ = '\\';		/* \\ is a \	*/
			rp += 1;
			}
		else
			*wp++ = *rp;
	while (*rp++ != '\0');

	return g_strdup(buf);
	}

static void
get_etc_fstab_list()
	{
	FILE			*f;
	Mount			*m;
	gchar			buf[512], *s;
	gchar			*dev, *dir, *type, *opt;

	while (fstab_list)
		{
		m = (Mount *) fstab_list->data;
		g_free(m->device);
		g_free(m->directory);
		g_free(m->type);
		g_free(m->options);
		g_free(m);
		fstab_list = g_list_remove(fstab_list, fstab_list->data);
		}
	if ((f = fopen("/etc/fstab", "r")) == NULL)
		return;
	while (fgets(buf, sizeof(buf), f))
		{
		s = buf;
		while (*s == ' ' || *s == '\t')
			++s;
		if (*s == '\0' || *s == '#' || *s == '\n')
			continue;
		dev = strtok(buf, " \t\n");
		dir = strtok(NULL, " \t\n");
		type = strtok(NULL, " \t\n");
		opt = strtok(NULL, " \t\n");

		if (   type == NULL
			|| strcmp(type, "devpts") == 0
			|| strcmp(type, "swap") == 0
			|| strcmp(type, "proc") == 0
			|| strcmp(type, "ignore") == 0
		   )
			continue;
		m = g_new0(Mount, 1);
		m->device = fix_fstab_name(dev);
		m->directory = fix_fstab_name(dir);
		m->type = fix_fstab_name(type);
		m->options =  fix_fstab_name(opt);

		fstab_list = g_list_append(fstab_list, m);
		}
	fclose(f);
	}


/* ----- Others -------------------------------------------------------- */
#if defined(USE_LIBGTOP)

#include <glibtop/mountlist.h>
#include <glibtop/fsusage.h>


static void
get_glibtop_mountlist()
	{
	glibtop_mountlist	mount_list;
	glibtop_mountentry	*mount_entries;
	Mount				*m;
	gint				i;

	while (proc_mounts_list)
		{
		m = (Mount *) proc_mounts_list->data;
		g_free(m->directory);
		g_free(m->device);
		g_free(m->type);
		g_free(proc_mounts_list->data);
		proc_mounts_list =
				g_list_remove(proc_mounts_list, proc_mounts_list->data);
		}
	mount_entries = glibtop_get_mountlist (&mount_list, 1);
	for (i = 0; i < mount_list.number; ++i)
		{
		m = g_new0(Mount, 1);
		m->directory = g_strdup(mount_entries[i].mountdir);
		m->device = g_strdup(mount_entries[i].devname);
		m->type = g_strdup(mount_entries[i].type);

		proc_mounts_list = g_list_append(proc_mounts_list, m);		
		}
	glibtop_free(mount_entries);
	}

static void
get_glibtop_fsusage(FSmon *fs)
	{
	glibtop_fsusage	fsusage;

	glibtop_get_fsusage(&fsusage, fs->mount.directory);
	fs->blocks = fsusage.blocks;
	fs->bavail = fsusage.bavail;
	fs->bfree = fsusage.bfree;
	fs->bsize  = 512;
	}
#endif


/* ----- Pick a system interface ----------------------------------------- */
static gint
setup_fs_interface()
    {
	uid = getuid();	/* only real root is allowed to mount/umount always */

#if defined(__FreeBSD__) || defined(__NetBSD__) || defined(__OpenBSD__)
    get_mountlist = get_bsd_mounts_list;
    get_fsusage = get_statfs_fsusage;
    get_fstab_list = get_etc_fstab_list;
#else
#if defined(__linux__)
	get_mountlist = get_linux_proc_mounts_list;
	get_fsusage = get_statfs_fsusage;
	get_fstab_list = get_etc_fstab_list;
#else
	get_mountlist = get_glibtop_mountlist;;
	get_fsusage = get_glibtop_fsusage;
	get_fstab_list = get_etc_fstab_list;
#endif
#endif
    return TRUE;
    }

/* ======================================================================== */

#define	MT_NONE		0
#define	MT_FSTAB	1
#define	MT_CUSTOM	2

#define	DEFAULT_DATA_FORMAT	_("$t - $f free")

  /* Values for force_fs_check	*/
#define	FORCE_REDRAW	1
#define	FORCE_UPDATE	2

static GtkWidget	*fs_main_vbox,
					*fs_drawer_vbox;

static gboolean		fs_check_timeout	= 2,
					nfs_check_timeout	= 16;
static gint			check_tick;

static gint			fs_drawer_is_shown;

static gint			n_fs_monitors;
static gint			force_fs_check;
static FSmon		*fs_in_motion;
static gint			x_fs_motion;
static gint			x_moved;

static gint			x_scroll;
static gint			text_decal_width;
static gint			cdrom_auto_eject;
static gint			binary_units;
static gchar		*data_format;

static gint			style_id;

gchar	*remote_fs_types[]	=
	{
	"nfs",
	"smbfs"
	};


static Mount *
in_fstab_list(gchar *s)
	{
	GList	*list;
	Mount	*m;

	for (list = fstab_list; list; list = list->next)
		{
		m = (Mount *)list->data;
		if (strcmp(s, m->directory) == 0)
			return m;
		}
	return NULL;
	}


static gint
fs_is_mounted(FSmon *fs)
	{
	Mount	*m_fs, *m_mounted;
	GList	*list;
	gint	i;

	fs->is_mounted = FALSE;
	m_fs = &fs->mount;
	for (list = proc_mounts_list; list; list = list->next)
		{
		m_mounted = (Mount *) list->data;
		if (strcmp(m_fs->directory, m_mounted->directory) == 0)
			{
			fs->is_mounted = TRUE;
			fs->is_nfs_fs = FALSE;
			for (i = 0; i < (sizeof(remote_fs_types) / sizeof(gchar *)); ++i)
				{
				if (!strcmp(m_mounted->type, remote_fs_types[i]))
					{
					fs->is_nfs_fs = TRUE;
					break;
					}
				}
			}
		}
	return fs->is_mounted;
	}

#if 0
#define MEGABYTES(x)		((gfloat)(x) * (binary_units \
			? (1024.0 * 1024.0) : 1e6))
#define GIGABYTES(x)		((gfloat)(x) * (binary_units \
			? (1024.0 * 1024.0 * 1024.0) : 1e9))
#define TERABYTES(x)		((gfloat)(x) * (binary_units \
			? (1024.0 * 1024.0 * 1024.0 * 1024.0) : 1e12))


static void
format_blocks(FSmon *fs, gchar *buf, gulong blocks)
	{
	gfloat			bytes;

	bytes = (gfloat) blocks * (gfloat) fs->bsize;

	if (bytes < MEGABYTES(10) )	/* Ramdisk, floppy */
		snprintf(buf, 32, "%.2fM", bytes / MEGABYTES(1) );
	else if (bytes < GIGABYTES(.9995) )
		snprintf(buf, 32, "%.0fM", bytes / MEGABYTES(1));
	else if (bytes < GIGABYTES(10))
		snprintf(buf, 32, "%.2fG", bytes / GIGABYTES(1));
	else if (bytes < GIGABYTES(20))
		snprintf(buf, 32, "%.1fG", bytes / GIGABYTES(1));
	else if (bytes < GIGABYTES(1000))
		snprintf(buf, 32, "%.0fG", bytes / GIGABYTES(1));
	else
		snprintf(buf, 32, "%.1fT", bytes / TERABYTES(1));
	}
#endif

static size_abbrev_table	fs_decimal_abbrev[] =
	{
	{ MB_SIZE(10),		MB_SIZE(1),		"%.2fM" },
	{ GB_SIZE(1),		MB_SIZE(1),		"%.0fM" },
	{ GB_SIZE(10),		GB_SIZE(1),		"%.2fG" },
	{ GB_SIZE(100),		GB_SIZE(1),		"%.1fG" },
	{ TB_SIZE(1),		GB_SIZE(1),		"%.0fG" },
	{ TB_SIZE(10),		TB_SIZE(1),		"%.2fT" },
	{ TB_SIZE(100),		TB_SIZE(1),		"%.1fT" }
	};

static size_abbrev_table	fs_binary_abbrev[] =
	{
	{ MiB_SIZE(10),		MiB_SIZE(1),	"%.2fM" },
	{ GiB_SIZE(1),		MiB_SIZE(1),	"%.0fM" },
	{ GiB_SIZE(10),		GiB_SIZE(1),	"%.2fG" },
	{ GiB_SIZE(100),	GiB_SIZE(1),	"%.1fG" },
	{ TiB_SIZE(1),		GiB_SIZE(1),	"%.0fG" },
	{ TiB_SIZE(10),		TiB_SIZE(1),	"%.2fT" },
	{ TiB_SIZE(100),	TiB_SIZE(1),	"%.1fT" }
	};

static gint
format_fs_data(FSmon *fs, gchar *buf, gint size)
	{
	gulong	b, u, f;
	gint	len;
	gchar	*s;
	gchar	tbuf[32], ubuf[32], fbuf[32];
	gfloat	bsize;
	size_abbrev_table	*tbl;
	size_t				tbl_size;

	--size;
	*buf = '\0';
	b = fs->blocks;
	u = fs->blocks - fs->bavail;
	f = fs->bavail;
	bsize = (gfloat) fs->bsize;

	tbl = binary_units ? &fs_binary_abbrev[0] : &fs_decimal_abbrev[0];
	tbl_size = binary_units
			? (sizeof(fs_binary_abbrev) / sizeof(size_abbrev_table))
			: (sizeof(fs_decimal_abbrev) / sizeof(size_abbrev_table));

	format_size_abbrev(tbuf, sizeof(tbuf), (gfloat)b * bsize, tbl, tbl_size);
	format_size_abbrev(ubuf, sizeof(ubuf), (gfloat)u * bsize, tbl, tbl_size);
	format_size_abbrev(fbuf, sizeof(fbuf), (gfloat)f * bsize, tbl, tbl_size);

	f = b - u;
	for (s = data_format; *s != '\0' && size > 0; ++s)
		{
		len = 1;
		if (*s == '$' && *(s + 1) != '\0')
			{
			switch(*(s + 1))
				{
				case 'l':
					len = snprintf(buf, size, "%s", fs->label);
					break;
				case 't':
					len = snprintf(buf, size, "%s", tbuf);
					break;
				case 'u':
					len = snprintf(buf, size, "%s", ubuf);
					break;
				case 'U':
					if (b > 0)
						len = snprintf(buf, size, "%ld%%",
							100 * (u >> 8) / (b >> 8));
					break;
				case 'f':
					len = snprintf(buf, size, "%s", fbuf);
					break;
				case 'F':
					if (b > 0)
						len = snprintf(buf, size, "%ld%%",
							100 * (f >> 8) / (b >> 8));
					break;
				default:
					*buf = *s;
					if (size > 1)
						{
						*(buf + 1) = *(s + 1);
						++len;
						}
					break;
				}
			++s;
			}
		else
			*buf = *s;
		size -= len;
		buf += len;
		}
	*buf = '\0';
	return u;
	}

  /* Draw the fs label or toggle the fs total blocks and blocks avail.
  */
static gint
draw_fs_text_decal(FSmon *fs, gint value)
	{
	Decal		*d;
	TextStyle	ts_save;
	gchar		buf[128];
	gint		changed;	/* Decal value, tells whether redraw required */
	gint		w	= 0;

	d = fs->text_decal;

	if (value == 0)
		{
		d->x_off = 0;
		gkrellm_draw_decal_text(&fs->panel, d, fs->label, 0);
		}
	else
		{
		ts_save = d->text_style;
		d->text_style = *gkrellm_meter_alt_textstyle(style_id);

		changed = format_fs_data(fs, buf, sizeof(buf));
		w = gdk_string_width(d->text_style.font, buf);
		if (w  > d->w)
			{
			d->x_off = d->w / 3 - x_scroll;
			changed = x_scroll + 1;
			}
		else
			d->x_off = 0;

		/* Draw the decal, actual draws occur only if "changed" is different
		|  from last call.  If scrolling, will be so each time, if not
		|  will be so as format_fs_data() return value changes.
		*/
		gkrellm_draw_decal_text(&fs->panel, d, buf, changed);
		d->text_style = ts_save;
		}
	return w;
	}

  /* Try to eject~ CDROMs.  If not a cdrom, the ioctl will just fail.
  |  Can't do zip or jazz without getting into SCSI commands.  Those
  |  could try a umount ; eject combo as a custom unmount command.
  */
static void
eject_cdrom(FSmon *fs)
	{
#if defined(CDROMEJECT) || defined(__FreeBSD__)
	Mount	*m;
	gint	d, n;

	m = in_fstab_list(fs->mount.directory);
	if (m != NULL)
		{
#if defined(__FreeBSD__)
		if ((d = open(m->device, O_RDONLY)) >= 0)
#else
		if ((d = open(m->device, (O_RDONLY | O_NONBLOCK))) >= 0)
#endif
			{
			extern int errno;

#if defined(__FreeBSD__)
			(void) ioctl(d, CDIOCALLOW);
			n = ioctl(d, CDIOCEJECT);
#else
			n = ioctl(d, CDROMEJECT);
#endif
			if (GK.debug_level & DEBUG_MISC)
				printf(_("fs eject %s errno = %d\n"), m->device, errno);
			close(d);
			}
		else if (GK.debug_level & DEBUG_MISC)
			printf(_("could not open %s\n"), m->device);
		}
#endif
	}

static void
accumulate_pipe_gstring(FSmon *fs)
	{
	gchar	buf[512];
	gint	n;

	n = fread(buf, 1, sizeof(buf) - 1, fs->launch_mount.pipe);
	buf[n] = '\0';
	if (n > 0)
		{
		if (fs->pipe_gstring)
			g_string_append(fs->pipe_gstring, buf);
		else
			fs->pipe_gstring = g_string_new(buf);
		}
	if (feof(fs->launch_mount.pipe))
		{
		pclose(fs->launch_mount.pipe);
		fs->launch_mount.pipe = NULL;
		}
	}

static void
pipe_command(FSmon *fs, gchar *command)
	{
	gchar	buf[512];

	if (fs->launch_mount.pipe)	/* Still running? */
		return;
	snprintf(buf, sizeof(buf), "%s 2>&1", command);
	if (GK.debug_level & DEBUG_MISC)
		printf(_("pipe mount command: %s\n"), buf);
	if ((fs->launch_mount.pipe = popen(buf, "r")) == NULL)
		return;
	fcntl(fileno(fs->launch_mount.pipe), F_SETFL, O_NONBLOCK);
	}

static void
mount_command(FSmon *fs)
	{
	gchar	cmd[CFG_BUFSIZE];

	if (fs->enable_mounting == MT_NONE)
		return;
	if (fs->is_mounted)
		{
		if (fs->enable_mounting == MT_FSTAB)
			snprintf(cmd, sizeof(cmd), "umount %s", fs->mount.directory);
		else	/* MT_CUSTOM */
			snprintf(cmd, sizeof(cmd), "%s", fs->launch_umount.command);
		fs->label_is_data = FALSE;
		draw_fs_text_decal(fs, 0);
		pipe_command(fs, cmd);
		if (cdrom_auto_eject)
			fs->eject_pending = GK.timer_ticks + 5;	/* at least 1/2 sec delay*/
		}
	else
		{
		if (fs->enable_mounting == MT_FSTAB)
			snprintf(cmd, sizeof(cmd), "mount %s", fs->mount.directory);
		else	/* MT_CUSTOM */
			snprintf(cmd, sizeof(cmd), "%s", fs->launch_mount.command);
		pipe_command(fs, cmd);
		}
	force_fs_check = FORCE_REDRAW;	/* An update triggers when pipe closes */
	}

static void
hide_fs_monitor(FSmon *fs)
	{
	if (fs->is_visible && !(fs_is_mounted(fs) && fs->show_if_mounted))
		{
		gtk_widget_hide(fs->vbox);
		gkrellm_monitor_height_adjust( - fs->panel.h);
		fs->is_visible = FALSE;
		}
	}

static void
hide_monitors_in_drawer()
	{
	FSmon	*fs;
	GList	*list;

	for (list = fs_mon_list; list; list = list->next)
		{
		fs = (FSmon *) list->data;
		if (fs->in_drawer)
			hide_fs_monitor(fs);
		}
	gkrellm_pack_side_frames();
	}

static void
show_fs_monitor(FSmon *fs)
	{
	if (!fs->is_visible)
		{
		gtk_widget_show(fs->vbox);
		gkrellm_monitor_height_adjust(fs->panel.h);
		fs->is_visible = TRUE;
		}
	}

static void
show_monitors_in_drawer()
	{
	FSmon	*fs;
	GList	*list;

	for (list = fs_mon_list; list; list = list->next)
		{
		fs = (FSmon *) list->data;
		if (fs->in_drawer)
			show_fs_monitor(fs);
		}
	gkrellm_pack_side_frames();
	}

static void
update_fs(void)
	{
	FSmon	*fs;
	Panel	*p;
	GList	*list;
	gulong	l;
	gint	index, w_scroll, w;
	gint	fs_check, nfs_check, next_check;

	if (fs_mon_list == NULL)
		return;

	w = w_scroll = 0;
	for (list = fs_mon_list; list; list = list->next)
		{
		fs = (FSmon *) list->data;
		if (fs->label_is_data && !fs_in_motion)
			{
			w = draw_fs_text_decal(fs, 1);
			if (w > w_scroll)
				w_scroll = w;
			gkrellm_draw_layers(&fs->panel);
			}
		}
	if (!fs_in_motion)
		{
		if (w_scroll > text_decal_width)
			x_scroll = (x_scroll + ((UC.update_HZ < 7) ? 2 : 1))
					% (w_scroll - text_decal_width / 3);
		else
			x_scroll = 0;
		}
	if (GK.second_tick)
		++check_tick;
	fs_check = (check_tick % fs_check_timeout) ? FALSE : TRUE;
	nfs_check = (check_tick % nfs_check_timeout) ? FALSE : TRUE;

	if (!force_fs_check && (!GK.second_tick || (!fs_check && !nfs_check)))
		return;
#if 0
printf("fs update %d nfs %d force %d\n", fs_check, nfs_check, force_fs_check);
#endif

	(*get_mountlist)();

	next_check = 0;

	for (list = fs_mon_list; list; list = list->next)
		{
		fs = (FSmon *) list->data;
		p  = &fs->panel;
		KRELL(p)->previous = 0;
		if (fs_is_mounted(fs))
			{
			if (fs->enable_mounting)
				{	/* Blink it while pipe is open.	*/
				if (   fs->launch_mount.pipe
					&& fs->md_button->cur_index == D_MISC_FS_MOUNTED
				   )
					index = D_MISC_FS_UMOUNTED;
				else
					index = D_MISC_FS_MOUNTED;
				gkrellm_set_decal_button_index(fs->md_button, index);
				}
			else
				gkrellm_set_decal_button_index(fs->md_button, D_MISC_LED1);
			if (   force_fs_check == FORCE_UPDATE
				|| (fs_check && !fs->is_nfs_fs)
				|| (nfs_check && fs->is_nfs_fs)
			   )
				(*get_fsusage)(fs);
			KRELL(p)->full_scale = fs->blocks;
			l = fs->blocks - fs->bavail;
			if (fs_in_motion && fs->label_is_data && x_moved)
				l = 0;
			gkrellm_update_krell(p, KRELL(p), l);
			if (fs->in_drawer && !fs->is_visible && fs->show_if_mounted)
				{
				show_fs_monitor(fs);
				gkrellm_pack_side_frames();
				}
			}
		else
			{
			if (fs->enable_mounting)
				{	/* Blink it while pipe is open.	*/
				if (   fs->launch_mount.pipe
					&& fs->md_button->cur_index == D_MISC_FS_UMOUNTED
				   )
					index = D_MISC_FS_MOUNTED;
				else
					index = D_MISC_FS_UMOUNTED;
				gkrellm_set_decal_button_index(fs->md_button, index);
				}
			else
				gkrellm_set_decal_button_index(fs->md_button, D_MISC_LED0);
			KRELL(p)->full_scale = 100;		/* Arbitrary > 0 */
			gkrellm_update_krell(p, KRELL(p), 0);
			if (!fs_drawer_is_shown && fs->in_drawer && fs->is_visible)
				{
				hide_fs_monitor(fs);
				gkrellm_pack_side_frames();
				}
			}
		fs->md_button->sensitive = fs->enable_mounting ? TRUE : FALSE;
		if (!fs->label_is_data && fs != fs_in_motion)
			draw_fs_text_decal(fs, 0);

		gkrellm_draw_layers(p);
		if (fs->eject_pending && fs->eject_pending < GK.timer_ticks)
			{
			eject_cdrom(fs);
			fs->eject_pending = 0;
			}
		if (fs->launch_mount.pipe)
			{
			accumulate_pipe_gstring(fs);
			if (fs->launch_mount.pipe == NULL)	/* Command is done */
				{
				if (fs->pipe_gstring)
					{
					gkrellm_message_window(_("GKrellM Mount Error"),
							fs->pipe_gstring->str, fs->panel.drawing_area);
					g_string_free(fs->pipe_gstring, 1);
					fs->pipe_gstring = NULL;
					}
				next_check = FORCE_UPDATE;
				}
			else
				next_check = FORCE_REDRAW; 	/* Keep it going */
			}
		}
	force_fs_check = next_check;
	}


  /* Real pixmaps are always in chart_minute even if chart_hour is displayed.
  */
static gint
fs_expose_event(GtkWidget *widget, GdkEventExpose *ev)
	{
	FSmon	*fs;
	GList	*list;

	for (list = fs_mon_list; list; list = list->next)
		{
		fs = (FSmon *) list->data;
		if (widget == fs->panel.drawing_area)
			{
			gdk_draw_pixmap(widget->window, GK.draw1_GC, fs->panel.pixmap,
					ev->area.x, ev->area.y, ev->area.x, ev->area.y,
					ev->area.width, ev->area.height);
			break;
			}
		}
	return FALSE;
	}

static gint
cb_fs_panel_release(GtkWidget *widget, GdkEventButton *ev)
	{
	if (ev->button == 1)
		return TRUE;
	if (fs_in_motion /* && fs_in_motion->panel.drawing_area == widget */)
		{
		if (fs_in_motion->restore_label)
			{
			if (fs_in_motion->label_is_data)
				gkrellm_config_modified();
			fs_in_motion->label_is_data = FALSE;
			draw_fs_text_decal(fs_in_motion, 0);
			gkrellm_draw_layers(&fs_in_motion->panel);
			}
		fs_in_motion->restore_label = TRUE;
		}
	force_fs_check = FORCE_REDRAW;		/* Move krells back */
	fs_in_motion = NULL;
	x_moved = 0;
	return TRUE;
	}

static gint
cb_fs_panel_press(GtkWidget *widget, GdkEventButton *ev)
	{
	FSmon	*fs;
	GList	*list;
	Decal	*d;

	for (list = fs_mon_list; list; list = list->next)
		{
		fs = (FSmon *) list->data;
		if (widget != fs->panel.drawing_area)
			continue;
		d = fs->mount_decal;
		if (ev->button == 1 && (ev->x < d->x || ev->x > d->x + d->w))
			{
			if (fs_drawer_is_shown)
				hide_monitors_in_drawer();
			else
				show_monitors_in_drawer();
			fs_drawer_is_shown = 1 - fs_drawer_is_shown;
			}
		else if (fs->is_mounted && (ev->button == 2 || ev->button == 3))
			{
			if (!fs->label_is_data)
				{
				fs->label_is_data = TRUE;
				fs->restore_label = FALSE;
				gkrellm_config_modified();
				}
			x_fs_motion = ev->x;
			draw_fs_text_decal(fs, 1);
			gkrellm_draw_layers(&fs->panel);
			fs_in_motion = fs;
			x_moved = 0;
			}
		break;
		}
	return TRUE;
	}

static gint
cb_fs_panel_motion(GtkWidget *widget, GdkEventButton *ev)
	{
	GdkModifierType	state;
	GList			*list;
	FSmon			*fs;
	Decal			*d;
	GdkFont			*font;
	gchar			buf[128];
	gint			w, x_delta	= 0;

	state = ev->state;
	if (   !fs_in_motion
		|| !(state & (GDK_BUTTON2_MASK | GDK_BUTTON3_MASK))
		|| !fs_in_motion->label_is_data
	   )
		{
		fs_in_motion = NULL;
		return TRUE;
		}
	font = gkrellm_meter_alt_textstyle(style_id)->font;

	d = fs_in_motion->text_decal;
	format_fs_data(fs_in_motion, buf, sizeof(buf));
	w = gdk_string_width(font, buf);
	if (w > d->w)
		{
		x_delta = ev->x - x_fs_motion;
		x_fs_motion = ev->x;
		d->x_off += x_delta;
		if (d->x_off < -w)
			d->x_off = -w;
		if (d->x_off > d->w)
			d->x_off = d->w;
		x_scroll = d->w / 3 - d->x_off;
		for (list = fs_mon_list; list; list = list->next)
			{
			fs = (FSmon *) list->data;
			if (fs->label_is_data)
				{
				draw_fs_text_decal(fs, 1);
				gkrellm_draw_layers(&fs->panel);
				}
			}
		if (x_moved > 0)
			fs_in_motion->restore_label = FALSE;
		}
	if (x_moved == 0)
		force_fs_check = FORCE_REDRAW;	/* Move krells out of the way */
	x_moved += (x_delta > 0) ? x_delta : -x_delta;
	return TRUE;
	}

  /* GKrellM button code in krell.c calls this when the mount_decal_button
  |  gets a press and release.
  */
static void
cb_fs_mount_button(DecalButton *button)
	{
	if (button)
		mount_command((FSmon *) button->data);
	}


static void
create_fs_monitor(GtkWidget *vbox1, FSmon *fs, gint index, gint first_create)
	{
	GtkWidget	*vbox;
	Style		*style;
	Panel		*p;
	gint		x, w;

	p = &fs->panel;
	if (first_create)
		{
		vbox = gtk_vbox_new(FALSE, 0);
		gtk_container_add(GTK_CONTAINER(vbox1), vbox);
		fs->vbox = vbox;

		p->textstyle = gkrellm_textstyle_new0();
		p->label = gkrellm_label_new0();
		}
	else		/* Theme or size change, will recreate decals and krells */
		{
		vbox = fs->vbox;
		gkrellm_destroy_decal_list(p);
		gkrellm_destroy_krell_list(p);
		}

	fs->idx = index;
	++n_fs_monitors;

	style = gkrellm_meter_style(style_id);
	p->textstyle = gkrellm_meter_textstyle(style_id);
	x = style->margin;

	fs->text_decal = gkrellm_create_decal_text(p, "Ap0", p->textstyle,
				style, x, -1, -1);	/* -1 means use y default */

	fs->mount_decal = gkrellm_create_decal_pixmap(p, GK.decal_misc_pixmap,
			GK.decal_misc_mask, N_MISC_DECALS, style, -1, -1);

	/* The default text decal has width of chart_width - panel borders and
	|  x position at left border, so calculate the override values.
	|  Cannot override w after the create.
	*/
	if (style->label_position >= 50)
		x += fs->mount_decal->w;
	w = UC.chart_width - fs->mount_decal->w
			- 2 * style->margin;
	text_decal_width = w;

	gkrellm_create_krell(p, gkrellm_krell_meter_image(style_id), style);

	gkrellm_configure_panel(p, NULL, style);

	gkrellm_create_panel(vbox, p, gkrellm_bg_meter_image(style_id));

	/* Center the mount decal
	*/
	fs->mount_decal->y = (p->h - fs->mount_decal->h) / 2;

	fs->md_button = gkrellm_make_decal_button(p, fs->mount_decal,
			cb_fs_mount_button, fs, D_MISC_FS_UMOUNTED, D_MISC_FS_PRESSED);

	if (first_create)
		{
		gtk_signal_connect(GTK_OBJECT(p->drawing_area), "expose_event",
				(GtkSignalFunc) fs_expose_event, NULL);
		gtk_signal_connect(GTK_OBJECT(p->drawing_area),"button_press_event",
				(GtkSignalFunc) cb_fs_panel_press, NULL);
		gtk_signal_connect(GTK_OBJECT(p->drawing_area),"button_release_event",
				(GtkSignalFunc) cb_fs_panel_release, NULL);
		gtk_signal_connect(GTK_OBJECT(p->drawing_area),"motion_notify_event",
				(GtkSignalFunc) cb_fs_panel_motion, NULL);
		if (!fs->in_drawer || (fs_is_mounted(fs) && fs->show_if_mounted))
			show_fs_monitor(fs);
		}
	else if (fs->is_visible)
		gkrellm_monitor_height_adjust(p->h);

	draw_fs_text_decal(fs, 0);
	force_fs_check = FORCE_UPDATE;

	if (fs->launch_mount.command == NULL)
		fs->launch_mount.command = g_strdup("");
	if (fs->launch_umount.command == NULL)
		fs->launch_umount.command = g_strdup("");
	}

static void
destroy_fs_monitor(FSmon *fs)
	{
	g_free(fs->label);
	g_free(fs->mount.directory);
	if (fs->is_visible)
		gkrellm_monitor_height_adjust( - fs->panel.h);
	gkrellm_destroy_panel(&fs->panel);
	gtk_widget_destroy(fs->vbox);
	g_free(fs);
	--n_fs_monitors;
	}

static void
create_fs(GtkWidget *vbox, gint first_create)
	{
	GList		*list;
	FSmon		*fs;
	gint		i;

	/* If first time called, create the container widgets.  Subsequent
	|  calls will be for theme or size changes.
	*/
	if (fs_main_vbox == NULL)
		{
		fs_main_vbox = gtk_vbox_new(FALSE, 0);
		gtk_container_add(GTK_CONTAINER(vbox), fs_main_vbox);
		gtk_widget_show(fs_main_vbox);

		fs_drawer_vbox = gtk_vbox_new(FALSE, 0);
		gtk_container_add(GTK_CONTAINER(vbox), fs_drawer_vbox);
		gtk_widget_show(fs_drawer_vbox);
		fs_drawer_is_shown = FALSE;
		}
	n_fs_monitors = 0;
	for (i = 0, list = fs_mon_list; list; ++i, list = list->next)
		{
		fs = (FSmon *)list->data;
		create_fs_monitor(fs->in_drawer ? fs_drawer_vbox : fs_main_vbox,fs,
					i, first_create);
		}
	}

#define	FS_CONFIG_KEYWORD	"fs"

static void
save_fs_config(FILE *f)
	{
	GList	*list;
	FSmon	*fs;

	for (list = fs_mon_list; list; list = list->next)
		{
		fs = (FSmon *) list->data;
		fprintf(f, "%s \"%s\" %s %d %d %d %d\n", FS_CONFIG_KEYWORD,
				fs->label, fs->mount.directory,
				fs->enable_mounting, fs->in_drawer,
				fs->show_if_mounted, fs->label_is_data);
		if (*(fs->launch_mount.command) != '\0')
			fprintf(f, "%s mount_command %s\n", FS_CONFIG_KEYWORD,
					fs->launch_mount.command);
		if (*(fs->launch_umount.command) != '\0')
			fprintf(f, "%s umount_command %s\n", FS_CONFIG_KEYWORD,
					fs->launch_umount.command);
		}
	fprintf(f, "%s fs_check_timeout %d\n", FS_CONFIG_KEYWORD,
			fs_check_timeout);
	fprintf(f, "%s nfs_check_timeout %d\n", FS_CONFIG_KEYWORD,
			nfs_check_timeout);
	fprintf(f, "%s auto_eject %d\n", FS_CONFIG_KEYWORD, cdrom_auto_eject);
	fprintf(f, "%s binary_units %d\n", FS_CONFIG_KEYWORD, binary_units);
	fprintf(f, "%s data_format %s\n", FS_CONFIG_KEYWORD, data_format);
	}

static gint
fix_fstab_mountable_changed(FSmon *fs)
	{
	Mount	*m;

	m = in_fstab_list(fs->mount.directory);
	if (   (!m || (!strstr(m->options, "user") && uid != 0))
		&& fs->enable_mounting == MT_FSTAB
	   )
		{
		fs->enable_mounting = MT_NONE;
		return TRUE;
		}		
	return FALSE;
	}

static void
load_fs_config(gchar *arg)
	{
	static FSmon	*fs_prev;
	FSmon			*fs;
	gchar			*label;
	gchar			config[32], item[CFG_BUFSIZE], dir[CFG_BUFSIZE];
	gint 			drawer, show, mnt, n;
	gint			is_data = 0;

	if ((n = sscanf(arg, "%31s %[^\n]", config, item)) != 2)
		return;

	if (!strcmp(config, "fs_check_timeout"))
		{
		sscanf(item, "%d", &fs_check_timeout);
		if (fs_check_timeout < 2)
			fs_check_timeout = 2;
		}
	else if (!strcmp(config, "nfs_check_timeout"))
		{
		sscanf(item, "%d", &nfs_check_timeout);
		if (nfs_check_timeout < 5)
			nfs_check_timeout = 5;
		}
	else if (!strcmp(config, "auto_eject"))
		sscanf(item, "%d", &cdrom_auto_eject);
	else if (!strcmp(config, "binary_units"))
		sscanf(item, "%d", &binary_units);
	else if (!strcmp(config, "data_format"))
		gkrellm_dup_string(&data_format, item);
	else if (fs_prev && !strcmp(config, "mount_command"))
		gkrellm_dup_string(&fs_prev->launch_mount.command, item);
	else if (fs_prev && !strcmp(config, "umount_command"))
		gkrellm_dup_string(&fs_prev->launch_umount.command, item);
	else
		{
		if ((label = cut_quoted_string(arg, &arg)) != NULL)
			{
			mnt = drawer = show = 0;
			dir[0] = '\0';
			sscanf(arg, "%s %d %d %d %d", dir, &mnt, &drawer, &show, &is_data);
			fs = g_new0(FSmon, 1);
			fs->label           = g_strdup(label);
			fs->mount.directory = g_strdup(dir);
			fs->enable_mounting = mnt;
			fix_fstab_mountable_changed(fs);
			fs->in_drawer = drawer;
			fs->show_if_mounted = show;
			if (is_data)
				{
				fs->label_is_data = TRUE;
				fs->restore_label = TRUE;
				}
			fs->launch_mount.command = g_strdup("");
			fs->launch_umount.command = g_strdup("");
			fs_mon_list = g_list_append(fs_mon_list, fs);
			fs_prev = fs;
			}
		}
	}


/* --------------------------------------------------------------------- */

static GtkWidget	*primary_label_entry,
					*primary_dir_combo,
					*secondary_label_entry,
					*secondary_dir_combo,
					*primary_mount_entry,
					*primary_umount_entry,
					*secondary_mount_entry,
					*secondary_umount_entry,
					*data_format_combo;

static GtkWidget	*fs_check_spin_button,
					*nfs_check_spin_button;

static GtkWidget	*primary_clist,
					*secondary_clist;

static GtkWidget	*primary_mounting_button,
					*secondary_mounting_button;
static GtkWidget	*show_if_mounted_button;
static GtkWidget	*cdrom_auto_eject_button,
					*binary_units_button;

static gint			primary_selected_row	= -1;
static gint			secondary_selected_row	= -1;

static gint			fs_list_modified;
static gint			primary_drawer,
					secondary_drawer;

static void
select_drawer_mount_entries(gpointer drawer, GtkWidget **mount_entry,
		GtkWidget **umount_entry)
	{
	if (drawer == &primary_drawer)
		{
		*mount_entry = primary_mount_entry;
		*umount_entry = primary_umount_entry;
		}
	else
		{
		*mount_entry = secondary_mount_entry;
		*umount_entry = secondary_umount_entry;
		}
	}

static void
select_drawer_widgets(gpointer drawer, GtkWidget **combo, GtkWidget **button1,
		GtkWidget **button2, GtkWidget **entry, GtkWidget **clist, gint **row)
	{
	if (drawer == &primary_drawer)
		{
		*combo = primary_dir_combo;
		*button1 = primary_mounting_button;
		if (button2)
			*button2 = NULL;
		if (entry)
			*entry = primary_label_entry;
		if (clist)
			*clist = primary_clist;
		if (row)
			*row = &primary_selected_row;
		}
	else
		{
		*combo = secondary_dir_combo;
		*button1 = secondary_mounting_button;
		if (button2)
			*button2 = show_if_mounted_button;
		if (entry)
			*entry = secondary_label_entry;
		if (clist)
			*clist = secondary_clist;
		if (row)
			*row = &secondary_selected_row;
		}
	}


  /* Watch what is going into the directory combo entry, compare it to
  |  fstab entries and accordingly set sensitivity of the mounting_button.
  */
static void
cb_combo_changed(GtkWidget *widget, gpointer drawer)
	{
	GtkWidget	*dir_combo,
				*mounting_button;
	GtkWidget	*mount_entry,
				*umount_entry;
	Mount		*m;
	gchar		*s;

	select_drawer_widgets(drawer, &dir_combo, &mounting_button, NULL, NULL,
				NULL, NULL);
	select_drawer_mount_entries(drawer, &mount_entry, &umount_entry);

	s = gkrellm_entry_get_text(&(GTK_COMBO(dir_combo)->entry));
	m = in_fstab_list(s);
	if (m != NULL && (strstr(m->options, "user") != NULL || uid == 0))
		{
		gtk_widget_set_sensitive(mounting_button, TRUE);
		if (GTK_TOGGLE_BUTTON(mounting_button)->active)
			{
			gtk_entry_set_text(GTK_ENTRY(mount_entry), "");
			gtk_entry_set_text(GTK_ENTRY(umount_entry), "");
			gtk_widget_set_sensitive(mount_entry, FALSE);
			gtk_widget_set_sensitive(umount_entry, FALSE);
			}
		}
	else
		{
		if (GTK_TOGGLE_BUTTON(mounting_button)->active)
			gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(mounting_button),
					FALSE);
		else
			{
			gtk_widget_set_sensitive(mount_entry, TRUE);
			gtk_widget_set_sensitive(umount_entry, TRUE);
			}
		gtk_widget_set_sensitive(mounting_button, FALSE);
		}
	}

static void
cb_mount_button_clicked(GtkWidget *widget, gpointer drawer)
	{
	GtkWidget	*mounting_button;
	GtkWidget	*mount_entry,
				*umount_entry;

	mounting_button = (drawer == &primary_drawer)
				? primary_mounting_button : secondary_mounting_button;
	select_drawer_mount_entries(drawer, &mount_entry, &umount_entry);

	if (GTK_TOGGLE_BUTTON(mounting_button)->active)
		{
		gtk_entry_set_text(GTK_ENTRY(mount_entry), "");
		gtk_entry_set_text(GTK_ENTRY(umount_entry), "");
		gtk_widget_set_sensitive(mount_entry, FALSE);
		gtk_widget_set_sensitive(umount_entry, FALSE);
		}
	else
		{
		gtk_widget_set_sensitive(mount_entry, TRUE);
		gtk_widget_set_sensitive(umount_entry, TRUE);
		}
	}

static void
reset_fs_entries(GtkWidget *entry, GtkWidget *combo, GtkWidget *button1,
		GtkWidget *button2, GtkWidget *mount_entry, GtkWidget *umount_entry)
	{
	gtk_entry_set_text(GTK_ENTRY(entry), "");
	gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(combo)->entry), "");
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button1), FALSE);
	if (button2)
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button2), FALSE);
	gtk_entry_set_text(GTK_ENTRY(mount_entry), "");
	gtk_entry_set_text(GTK_ENTRY(umount_entry), "");
	}

static void
cb_fs_clist_selected(GtkWidget *clist, gint row, gint column,
			GdkEventButton *bevent, gpointer drawer)
	{
	GtkWidget	*dir_combo,
				*label_entry,
				*mount_entry,
				*umount_entry,
				*mounting_button,
				*show_button;
	gchar		*s;
	gint		state, *selected_row, i;

	select_drawer_widgets(drawer, &dir_combo, &mounting_button, &show_button,
				&label_entry, NULL, &selected_row);
	select_drawer_mount_entries(drawer, &mount_entry, &umount_entry);

	i = 0;
	gtk_clist_get_text(GTK_CLIST(clist), row, i++, &s);
	gtk_entry_set_text(GTK_ENTRY(label_entry), s);

	gtk_clist_get_text(GTK_CLIST(clist), row, i++, &s);
	gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(dir_combo)->entry), s);

	if (show_button)
		{
		gtk_clist_get_text(GTK_CLIST(clist), row, i++, &s);
		state = (strcmp(s, "yes") == 0) ? TRUE : FALSE;
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(show_button), state);
		}

	gtk_clist_get_text(GTK_CLIST(clist), row, i++, &s);
	state = (strcmp(s, "yes") == 0) ? TRUE : FALSE;
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(mounting_button), state);

	gtk_clist_get_text(GTK_CLIST(clist), row, i++, &s);
	gtk_entry_set_text(GTK_ENTRY(mount_entry), s);
	gtk_clist_get_text(GTK_CLIST(clist), row, i++, &s);
	gtk_entry_set_text(GTK_ENTRY(umount_entry), s);

	*selected_row = row;
	}

static void
cb_fs_clist_unselected(GtkWidget *clist, gint row, gint column,
			GdkEventButton *bevent, gpointer drawer)
	{
	GtkWidget	*dir_combo,
				*label_entry,
				*mount_entry,
				*umount_entry,
				*mounting_button,
				*show_button;
	gint		*selected_row;

	select_drawer_widgets(drawer, &dir_combo, &mounting_button, &show_button,
				&label_entry, NULL, &selected_row);
	select_drawer_mount_entries(drawer, &mount_entry, &umount_entry);
	reset_fs_entries(label_entry, dir_combo, mounting_button, show_button,
			mount_entry, umount_entry);
	*selected_row = -1;
	}

static void
cb_fs_clist_up(GtkWidget *widget, gpointer drawer)
	{
	GtkWidget	*clist;
	gint		row, *selected_row;

	clist = (drawer == &primary_drawer) ? primary_clist : secondary_clist;
	selected_row = (drawer == &primary_drawer) ? &primary_selected_row
						: &secondary_selected_row;
	row = *selected_row;
	if (row > 0)
		{
		gtk_clist_row_move(GTK_CLIST(clist), row, row - 1);
		gtk_clist_select_row(GTK_CLIST(clist), row - 1, -1);
		if (gtk_clist_row_is_visible(GTK_CLIST(clist), row - 1)
					!= GTK_VISIBILITY_FULL)
			gtk_clist_moveto(GTK_CLIST(clist), row - 1, -1, 0.0, 0.0);
		*selected_row = row - 1;
		fs_list_modified = TRUE;
		}
	}

static void
cb_fs_clist_down(GtkWidget *widget, gpointer drawer)
	{
	GtkWidget	*clist;
	gint		row, *selected_row;

	clist = (drawer == &primary_drawer) ? primary_clist : secondary_clist;
	selected_row = (drawer == &primary_drawer) ? &primary_selected_row
						: &secondary_selected_row;
	row = *selected_row;
	if (row >= 0 && row < GTK_CLIST(clist)->rows - 1)
		{
		gtk_clist_row_move(GTK_CLIST(clist), row, row + 1);
		gtk_clist_select_row(GTK_CLIST(clist), row + 1, -1);
		if (gtk_clist_row_is_visible(GTK_CLIST(clist), row + 1)
					!= GTK_VISIBILITY_FULL)
			gtk_clist_moveto(GTK_CLIST(clist), row + 1, -1, 1.0, 0.0);
		*selected_row = row + 1;
		fs_list_modified = TRUE;
		}
	}

static void
cb_enter_fs(GtkWidget *widget, gpointer drawer)
	{
	GtkWidget	*dir_combo,
				*label_entry,
				*mount_entry,
				*umount_entry,
				*mounting_button,
				*show_button,
				*clist;
	gchar		*buf[7];
	gint		*selected_row, i, n;

	select_drawer_widgets(drawer, &dir_combo, &mounting_button, &show_button,
				&label_entry, &clist, &selected_row);
	select_drawer_mount_entries(drawer, &mount_entry, &umount_entry);
	i = 0;
	buf[i++] = gkrellm_entry_get_text(&label_entry);
	buf[i++] = gkrellm_entry_get_text(&(GTK_COMBO(dir_combo)->entry));
	if (show_button)
		buf[i++] = GTK_TOGGLE_BUTTON(show_button)->active ? "yes" : "no";
	buf[i++] = GTK_TOGGLE_BUTTON(mounting_button)->active ? "yes" : "no";
	n = i;
	buf[i++] = gkrellm_entry_get_text(&mount_entry);
	buf[i++] = gkrellm_entry_get_text(&umount_entry);
	buf[i] = NULL;
	if (*(buf[0]) == '\0' || *(buf[1]) == '\0')	/* validate we have input */
		{
		gkrellm_config_message_window(_("Entry Error"),
			_("Both a label and a mount point must be entered."), widget);
		return;
		}
	if ((*(buf[n]) && !*(buf[n+1])) || (!*(buf[n]) && *(buf[n+1])))
		{
		gkrellm_config_message_window(_("Entry Error"),
			_("Both mount and umount commands must be entered."), widget);
		return;
		}
	if (*selected_row >= 0)
		{
		for (i = 0; i < (show_button ? 6 : 5); ++i)
			gtk_clist_set_text(GTK_CLIST(clist), *selected_row, i, buf[i]);
		gtk_clist_unselect_row(GTK_CLIST(clist), *selected_row, 0);
		*selected_row = -1;
		}
	else
		gtk_clist_append(GTK_CLIST(clist), buf);
	reset_fs_entries(label_entry, dir_combo, mounting_button, show_button,
			mount_entry, umount_entry);
	fs_list_modified = TRUE;
	}


static void
cb_delete_fs(GtkWidget *widget, gpointer drawer)
	{
	GtkWidget	*dir_combo,
				*label_entry,
				*mount_entry,
				*umount_entry,
				*mounting_button,
				*show_button,
				*clist;
	gint		*selected_row;
	
	select_drawer_widgets(drawer, &dir_combo, &mounting_button, &show_button,
				&label_entry, &clist, &selected_row);
	select_drawer_mount_entries(drawer, &mount_entry, &umount_entry);
	reset_fs_entries(label_entry, dir_combo, mounting_button, show_button,
				mount_entry, umount_entry);
	if (*selected_row >= 0)
		{
		gtk_clist_remove(GTK_CLIST(clist), *selected_row);
		fs_list_modified = TRUE;
		*selected_row = -1;
		}
	}

  /* Transfer the state of the config tab clists to the fs_mon_list glist.
  |  There are separate primary and secondary clists while only one glist,
  |  so the clists are merged into a new glist which replaces the old
  |  fs_mon_list.  FSmon structs from the old fs_mon_list are reused if
  |  possible (overly complicated? - I could just destroy all then re-create).
  */
static void
apply_fs_config()
	{
	GList			*new_fs_list, *old_fs_list;
	FSmon			*fs;
	gchar			*s;
	gint			row, need_new_monitor;
	GtkSpinButton	*spin;

	cdrom_auto_eject = GTK_TOGGLE_BUTTON(cdrom_auto_eject_button)->active;
	binary_units = GTK_TOGGLE_BUTTON(binary_units_button)->active;
	s = gkrellm_entry_get_text(&(GTK_COMBO(data_format_combo)->entry));
	gkrellm_dup_string(&data_format, s);

	spin = GTK_SPIN_BUTTON(fs_check_spin_button);
	fs_check_timeout = gtk_spin_button_get_value_as_int(spin);
	spin = GTK_SPIN_BUTTON(nfs_check_spin_button);
	nfs_check_timeout = gtk_spin_button_get_value_as_int(spin);

	if (!fs_list_modified)
		return;
	new_fs_list = NULL;
	old_fs_list = fs_mon_list;

	/* Append primary_clist entries to the new_fs_list and reuse old primary
	|  structures as long as possible.
	*/
	for (row = 0; row < GTK_CLIST(primary_clist)->rows; ++row)
		{
		if (old_fs_list && ((FSmon *) old_fs_list->data)->in_drawer == FALSE)
			{
			fs = (FSmon *) old_fs_list->data;
			fs->text_decal->value = -1;
			old_fs_list = old_fs_list->next;
			need_new_monitor = FALSE;
			}
		else
			{
			fs = g_new0(FSmon, 1);
			need_new_monitor = TRUE;
			}
		gtk_clist_get_text(GTK_CLIST(primary_clist), row, 0, &s);
		gkrellm_dup_string(&fs->label, s);
		gtk_clist_get_text(GTK_CLIST(primary_clist), row, 1, &s);
		gkrellm_dup_string(&fs->mount.directory, s);

		gtk_clist_get_text(GTK_CLIST(primary_clist), row, 2, &s);
		fs->enable_mounting = (strcmp(s, "yes") == 0) ? MT_FSTAB : MT_NONE;
		gtk_clist_get_text(GTK_CLIST(primary_clist), row, 3, &s);
		gkrellm_dup_string(&fs->launch_mount.command, s);
		gtk_clist_get_text(GTK_CLIST(primary_clist), row, 4, &s);
		gkrellm_dup_string(&fs->launch_umount.command, s);
		if (*s |= '\0')
			fs->enable_mounting = MT_CUSTOM;

		fs->in_drawer = FALSE;
		new_fs_list = g_list_append(new_fs_list, fs);
		if (need_new_monitor)
			create_fs_monitor(fs_main_vbox, fs, n_fs_monitors, TRUE);
		}

	/* Destroy any remaining primary monitors left in the old fs_mon_list.
	*/
	while (old_fs_list && ((FSmon *) old_fs_list->data)->in_drawer == FALSE)
		{
		destroy_fs_monitor((FSmon *) old_fs_list->data);
		old_fs_list = old_fs_list->next;
		}

	/* Append the secondary clist monitors to the new_fs_list.
	*/
	for (row = 0; row < GTK_CLIST(secondary_clist)->rows; ++row)
		{
		if (old_fs_list && ((FSmon *) old_fs_list->data)->in_drawer == TRUE)
			{
			fs = (FSmon *) old_fs_list->data;
			fs->text_decal->value = -1;
			old_fs_list = old_fs_list->next;
			need_new_monitor = FALSE;
			}
		else
			{
			fs = g_new0(FSmon, 1);
			need_new_monitor = TRUE;
			}
		gtk_clist_get_text(GTK_CLIST(secondary_clist), row, 0, &s);
		gkrellm_dup_string(&fs->label, s);
		gtk_clist_get_text(GTK_CLIST(secondary_clist), row, 1, &s);
		gkrellm_dup_string(&fs->mount.directory, s);

		gtk_clist_get_text(GTK_CLIST(secondary_clist), row, 2, &s);
		fs->show_if_mounted = (strcmp(s, "yes") == 0) ? TRUE : FALSE;
		gtk_clist_get_text(GTK_CLIST(secondary_clist), row, 3, &s);
		fs->enable_mounting = (strcmp(s, "yes") == 0) ? MT_FSTAB : MT_NONE;
		gtk_clist_get_text(GTK_CLIST(secondary_clist), row, 4, &s);
		gkrellm_dup_string(&fs->launch_mount.command, s);
		gtk_clist_get_text(GTK_CLIST(secondary_clist), row, 5, &s);
		gkrellm_dup_string(&fs->launch_umount.command, s);
		if (*s |= '\0')
			fs->enable_mounting = MT_CUSTOM;

		fs->in_drawer = TRUE;

		new_fs_list = g_list_append(new_fs_list, fs);
		if (need_new_monitor)
			create_fs_monitor(fs_drawer_vbox, fs, n_fs_monitors, TRUE);
		else if (fs_drawer_is_shown)
			show_fs_monitor(fs);
		else
			hide_fs_monitor(fs);
		}
	/* Destroy any remaining secondary monitors left in the old fs_mon_list.
	*/
	while (old_fs_list && ((FSmon *) old_fs_list->data)->in_drawer == TRUE)
		{
		destroy_fs_monitor((FSmon *) old_fs_list->data);
		old_fs_list = old_fs_list->next;
		}
	/* And destroy the old fs_mon_list glist and reassign it to the new list.
	*/
	while (fs_mon_list)
		fs_mon_list = g_list_remove(fs_mon_list, fs_mon_list->data);
	fs_mon_list = new_fs_list;
	gkrellm_pack_side_frames();
	fs_list_modified = FALSE;
	force_fs_check = FORCE_UPDATE;
	}

static gchar *fs_title1[5] =
	{ N_("Label"), N_("Mount Point"), N_("fstab"),
		N_("mount Command"), N_("umount Command") };

static gchar *fs_title2[6] =
	{ N_("Label"), N_("Mount Point"), N_("Show"), N_("fstab"),
		N_("mount Command"), N_("umount Command") };

static void
fill_fs_tab(GtkWidget *vbox, GtkWidget **label_entry, GtkWidget **dir_combo,
			GtkWidget **mount_button, GtkWidget **show_button,
			GtkWidget **mount_entry, GtkWidget **umount_entry,
			GtkWidget **fs_clist, gpointer drawer)
	{
	GtkWidget		*table;
	GtkWidget		*hbox;
	GtkWidget		*label;
	GtkWidget		*button;
	GtkWidget		*arrow;
	GtkWidget		*scrolled;
	GList			*list, *combo_list;
	FSmon			*fs;
	gchar			*buf[7];
	gint			row, in_drawer, i, j;

	hbox = gtk_hbox_new(FALSE, 3);
	gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 2);

	label = gtk_label_new(_("Label"));
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 6);
	*label_entry = gtk_entry_new_with_max_length(16);
	gtk_widget_set_usize(*label_entry, show_button ? 60 : 100, 0);
	gtk_box_pack_start(GTK_BOX(hbox), *label_entry, FALSE, FALSE, 0);

	label = gtk_label_new(_("Mount Point"));
	gtk_box_pack_start(GTK_BOX(hbox), label, TRUE, TRUE, 6);
	*dir_combo = gtk_combo_new();
	gtk_box_pack_start(GTK_BOX(hbox), *dir_combo, FALSE, FALSE,0);
	combo_list = NULL;
	for (list = fstab_list; list; list = list->next)
		combo_list = g_list_append(combo_list,
					((Mount *)list->data)->directory);
	gtk_combo_set_popdown_strings( GTK_COMBO(*dir_combo), combo_list);
	gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(*dir_combo)->entry), "");
	gtk_signal_connect(GTK_OBJECT(GTK_COMBO(*dir_combo)->entry), "changed",
				GTK_SIGNAL_FUNC (cb_combo_changed), drawer);

	if (show_button)
		{
		*show_button = gtk_check_button_new_with_label(
						_("Show if mounted"));
		gtk_box_pack_start(GTK_BOX(hbox), *show_button, FALSE, FALSE, 4);
		}
	table = gkrellm_launcher_table_new(vbox, 1);	/* Won't have tooltip */
	gkrellm_config_launcher(table, 0, mount_entry, NULL, _("mount"), NULL);
	gkrellm_config_launcher(table, 1, umount_entry, NULL, _("umount"), NULL);

	hbox = gtk_hbox_new(FALSE, 3);
	gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 2);

	*mount_button = gtk_check_button_new_with_label(
					_("Enable /etc/fstab mounting"));
	gtk_box_pack_start(GTK_BOX(hbox), *mount_button, TRUE, TRUE, 0);
	gtk_signal_connect(GTK_OBJECT(GTK_BUTTON(*mount_button)), "clicked",
				GTK_SIGNAL_FUNC (cb_mount_button_clicked), drawer);

	button = gtk_button_new();
	arrow = gtk_arrow_new(GTK_ARROW_UP, GTK_SHADOW_ETCHED_OUT);
	gtk_container_add(GTK_CONTAINER(button), arrow);
	gtk_signal_connect(GTK_OBJECT(button), "clicked",
				(GtkSignalFunc) cb_fs_clist_up, drawer);
	gtk_box_pack_start(GTK_BOX(hbox), button, TRUE, TRUE, 4);

	button = gtk_button_new();
	arrow = gtk_arrow_new(GTK_ARROW_DOWN, GTK_SHADOW_ETCHED_OUT);
	gtk_container_add(GTK_CONTAINER(button), arrow);
	gtk_signal_connect(GTK_OBJECT(button), "clicked",
				(GtkSignalFunc) cb_fs_clist_down, drawer);
	gtk_box_pack_start(GTK_BOX(hbox), button, TRUE, TRUE, 4);

	button = gtk_button_new_with_label(_("Enter"));
	gtk_signal_connect(GTK_OBJECT(button), "clicked",
				(GtkSignalFunc) cb_enter_fs, drawer);
	gtk_box_pack_start(GTK_BOX(hbox), button, TRUE, TRUE, 4);

	button = gtk_button_new_with_label(_("Delete"));
	gtk_signal_connect(GTK_OBJECT(button), "clicked",
				(GtkSignalFunc) cb_delete_fs, drawer);
	gtk_box_pack_start(GTK_BOX(hbox), button, TRUE, TRUE, 4);

	scrolled = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolled),
			GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	gtk_box_pack_start(GTK_BOX(vbox), scrolled, TRUE, TRUE, 0);
	if (show_button)
		{
	    for(j = 0; j < 6; j++)
			fs_title2[j] = _(fs_title2[j]);
		*fs_clist = gtk_clist_new_with_titles(6, fs_title2);
		}
	else
		{
	    for(j = 0; j < 5; j++)
			fs_title1[j] = _(fs_title1[j]);
		*fs_clist = gtk_clist_new_with_titles(5, fs_title1);
		}
	gtk_clist_set_shadow_type (GTK_CLIST(*fs_clist), GTK_SHADOW_OUT);
	gtk_clist_set_column_width (GTK_CLIST(*fs_clist), 0, 60);
	gtk_clist_set_column_width (GTK_CLIST(*fs_clist), 1, 100);
	if (show_button)
		gtk_clist_set_column_width (GTK_CLIST(*fs_clist), 4, 200);
	else
		gtk_clist_set_column_width (GTK_CLIST(*fs_clist), 3, 200);

	gtk_signal_connect (GTK_OBJECT(*fs_clist), "select_row",
			(GtkSignalFunc) cb_fs_clist_selected, drawer);
	gtk_signal_connect (GTK_OBJECT(*fs_clist), "unselect_row",
			(GtkSignalFunc) cb_fs_clist_unselected, drawer);
	gtk_container_add (GTK_CONTAINER (scrolled), *fs_clist);
	in_drawer = (drawer == &secondary_drawer) ? TRUE : FALSE;
		
	for (list = fs_mon_list; list; list = list->next)
		{
		fs = (FSmon *) list->data;
		if (fs->in_drawer == in_drawer)
			{
			i = 0;
			buf[i++] = fs->label;
			buf[i++] = fs->mount.directory;
			if (show_button)
				buf[i++] = fs->show_if_mounted ? "yes" : "no";
			if (fix_fstab_mountable_changed(fs))
				fs_list_modified = TRUE;
			buf[i++] = (fs->enable_mounting == MT_FSTAB) ? "yes" : "no";
			buf[i++] = fs->launch_mount.command;
			buf[i++] = fs->launch_umount.command;
			buf[i] = NULL;
			row = gtk_clist_append(GTK_CLIST(*fs_clist), buf);
			}
		}
	}


static gchar *fs_info_text[] =
{
N_("Enter file system mount points to monitor and enter a label to describe\n"
"the mount point.  The krell shows the ratio of blocks used to total blocks\n"
"available.  Mounting commands can be enabled for mount points in one\n"
"of two ways:\n\n"),

N_("<b>\t1) Mounting using /etc/fstab: "),

N_("If a mount point is in your\n"
"\t/etc/fstab and you have mount permission then mount and umount\n"
"\tcommands can be enabled and executed for that mount point simply\n"
"\tby checking the \"Enable /etc/fstab mounting\" option.\n"
"\tMount table entries in /etc/fstab must have the \"user\" option\n"
"\tset to grant this permission unless GKrellM is run as root.\n"
"\tFor example, if you run GKrellM as a normal user and you want\n"
"\tto be able to mount your floppy, your /etc/fstab could have\n"
"\teither of:\n"),

N_("<i>\t\t/dev/fd0  /mnt/floppy  ext2   user,noauto,rw,exec  0  0\n"),
N_("\tor\n"),
N_("<i>\t\t/dev/fd0  /mnt/floppy  ext2   user,defaults  0  0\n\n"),

N_("<b>\t2) Mounting using custom commands: "),
N_("If GKrellM is run as root or if\n"
"\tyou have sudo permission to run the mount commands, then a custom\n"
"\tmount command can be entered into the \"mount command\" entry\n"
"\tbox.  A umount command must also be entered if you choose this\n"
"\tmethod.  Example mount and umount entries using sudo:\n"),

N_("<i>\t\tsudo /bin/mount -t msdos /dev/fd0 /mnt/A\n"
	"\t\tsudo /bin/umount /mnt/A\n"),

N_("\tNotes: the mount point specified in a custom mount command\n"
   "\t(/mnt/A in this example) must be the same as entered in the\n"
   "\t\"Mount Point\" entry.  You should have the NOPASSWD option set\n"
   "\tin /etc/sudoers if using sudo.\n\n"),

N_("<b>Primary and Secondary Monitors\n"),
N_("File system monitors can be created as primary (always visible)\n"
	"or secondary which can be hidden and then shown when they are of\n"
	"interest.  For example, you might make primary file system monitors\n"
	"for root, home, or user so they will be always visible, but make\n"
	"secondary monitors for less frequently used mount points such as\n"
	"floppy, zip, backup partitions, foreign file system types, etc.\n"
	"Secondary FS monitors can also be configured to always be visible if they\n"
	"are mounted by checking the \"Show if mounted\" option.   Using this\n"
	"feature you can show the secondary group, mount a file system, and have\n"
	"that FS monitor remain visible even when the secondary group is hidden.\n"
	"A standard cdrom mount will show as 100% full but a monitor for it\n"
	"could be created with mounting enabled just to have the\n"
	"mount/umount convenience.\n\n"),

N_("<b>Mouse Button Actions:\n"),
N_("<b>\tLeft "),
N_("click on a fs monitor label to show/hide secondary fs monitors.\n"),
N_("<b>\tLeft "),
N_("click on a drive decal to apply mount/umount commands (if enabled)\n"
	"\t\tto the mount point being monitored.\n"),
N_("<b>\tMiddle or Right "),
N_("click on a FS panel to scroll a programmable display\n"),
N_("\t\tof file system capacities (default is total and free space).\n\n"),

N_("At least one primary fs monitor must exist to click on in order to show\n"
"secondary ones.\n")
};

static void
create_fs_tab(GtkWidget *tab_vbox)
	{
	GtkWidget		*tabs;
	GtkWidget		*vbox, *hbox;
	GtkWidget		*scrolled;
	GtkWidget		*text;
	GtkWidget		*label;
	GtkWidget		*sep;
	GList			*list;
	gint			i;

	fs_list_modified = FALSE;
	(*get_fstab_list)();

	tabs = gtk_notebook_new();
	gtk_notebook_set_tab_pos(GTK_NOTEBOOK(tabs), GTK_POS_TOP);
	gtk_box_pack_start(GTK_BOX(tab_vbox), tabs, TRUE, TRUE, 0);

	vbox = gkrellm_create_tab(tabs, _("Primary"));
	fill_fs_tab(vbox, &primary_label_entry, &primary_dir_combo,
			&primary_mounting_button, NULL,
			&primary_mount_entry, &primary_umount_entry,
			&primary_clist, &primary_drawer);

	vbox = gkrellm_create_tab(tabs, _("Secondary"));
	fill_fs_tab(vbox, &secondary_label_entry, &secondary_dir_combo,
			&secondary_mounting_button, &show_if_mounted_button,
			&secondary_mount_entry, &secondary_umount_entry,
			&secondary_clist, &secondary_drawer);

/* --Options tab */
	vbox = gkrellm_create_tab(tabs, _("Options"));
	label = gtk_label_new(
_("A middle or right click in a File System monitor panel toggles a display\n"
"of capacities in a format controlled by a user supplied format string where\n"
"$t is total capacity, $u is used space, $f is free space, $U is used %,\n"
"$F is free %, and $l is the panel label."));
	gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, TRUE, 2);
	gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_LEFT);
	gtk_misc_set_alignment(GTK_MISC(label), 0, 0.5);

	hbox = gtk_hbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(vbox), hbox, TRUE, TRUE, 0);
	data_format_combo = gtk_combo_new();
	gtk_box_pack_start(GTK_BOX(hbox), data_format_combo, FALSE, TRUE, 0);
	list = NULL;
	list = g_list_append(list, DEFAULT_DATA_FORMAT);
	list = g_list_append(list, _("$t - $u used"));
	list = g_list_append(list, _("$t - $U"));
	gtk_combo_set_popdown_strings(GTK_COMBO(data_format_combo), list);
	gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(data_format_combo)->entry),
			data_format);
	label = gtk_label_new(_("Format string for FS data display"));
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, TRUE, 0);

	gkrellm_check_button(vbox, &binary_units_button, binary_units, FALSE, 4,
	_("Use binary units (MiB, GiG) for reporting disk capacities."));
	insert_expanded_filler(vbox);

	sep = gtk_hseparator_new();
	gtk_box_pack_start(GTK_BOX(vbox), sep, TRUE, TRUE, 0);

	gkrellm_check_button(vbox, &cdrom_auto_eject_button, cdrom_auto_eject,
				FALSE, 4, _("Enable eject when CD-ROMs are unmounted"));

	gkrellm_spin_button(vbox, &fs_check_spin_button, (gfloat) fs_check_timeout,
		1.0, 15.0, 1.0, 5.0, 0 /*digits*/, 0 /*width*/, NULL, NULL, FALSE,
		_("Seconds between data updates for local mounted file systems"));

	gkrellm_spin_button(vbox, &nfs_check_spin_button,
		(gfloat) nfs_check_timeout, 5.0, 60.0, 1.0, 5.0, 0, 0, NULL,NULL,FALSE,
		_("Seconds between data updates for remote mounted file systems"));

/* --Info tab */
	vbox = gkrellm_create_tab(tabs, _("Info"));
	scrolled = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolled),
			GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	gtk_box_pack_start(GTK_BOX(vbox), scrolled, TRUE, TRUE, 0);
	text = gtk_text_new(NULL, NULL);
	for (i = 0; i < sizeof(fs_info_text)/sizeof(gchar *); ++i)
		gkrellm_add_info_text_string(text, _(fs_info_text[i]));
	gtk_text_set_editable(GTK_TEXT(text), FALSE);
	gtk_container_add(GTK_CONTAINER(scrolled), text);
	}



static Monitor	monitor_fs =
	{
	N_("File System"),		/* Name, for config tab.	*/
	MON_FS,				/* Id,  0 if a plugin		*/
	create_fs,			/* The create function		*/
	update_fs,			/* The update function		*/
	create_fs_tab,		/* The config tab create function	*/
	apply_fs_config,	/* Apply the config function		*/

	save_fs_config,		/* Save user conifg			*/
	load_fs_config,		/* Load user config			*/
	FS_CONFIG_KEYWORD,	/* config keyword			*/

	NULL,				/* Undef 2	*/
	NULL,				/* Undef 1	*/
	NULL,				/* Undef 0	*/

	0,					/* insert_before_id - place plugin before this mon */

	NULL,				/* Handle if a plugin, filled in by GKrellM		*/
	NULL				/* path if a plugin, filled in by GKrellM		*/
	};

Monitor *
init_fs_monitor(void)
	{
	monitor_fs.name = _(monitor_fs.name);
	style_id = gkrellm_add_meter_style(&monitor_fs, FS_STYLE_NAME);
	data_format = g_strdup(DEFAULT_DATA_FORMAT);
	if (setup_fs_interface())
		{
		(*get_fstab_list)();
		(*get_mountlist)();
		return &monitor_fs;
		}
	return FALSE;
	}

