/* GKrellM
|  Copyright (C) 1999-2001 Bill Wilson
|
|  Author:  Bill Wilson    bill@gkrellm.net
|  Latest versions might be found at:  http://gkrellm.net
|
|  This program is free software which I release under the GNU General Public
|  License. You may redistribute and/or modify this program under the terms
|  of that license as published by the Free Software Foundation; either
|  version 2 of the License, or (at your option) any later version.
|
|  This program is distributed in the hope that it will be useful,
|  but WITHOUT ANY WARRANTY; without even the implied warranty of
|  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
|  GNU General Public License for more details.
| 
|  To get a copy of the GNU General Puplic License, write to the Free Software
|  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
| 1/14/2001 Patch from Masaharu FUJITA <fujita@hpc.cs.ehime-u.ac.jp>
|	The new mail krell animation for BSD is now a Daemon.
| 11/30/2000 Patch from Adam Sampson <ats1@ukc.ac.uk>
|	Added MH mail dir checking, check_mh_dir() 
| 10/19/2000 Patch from Matt McClure <mlm@aya.yale.edu>
|	Added tooltip reporting of new mail counts for mailboxes checked via the
|	Mail fetch/check program.
|  7/4/2000 Patch from Michael Bell <mikeb@map.com>
|	Added builtin POP3 and IMAP mail box checking routines.  Also added
|	reporting the new mail counts for these to the tooltip popup.
|	Hajimu UMEMOTO added IPv6 awareness to these routines.
|  4/25/2000 Patch from Stefan Schlott <stefan.schlott@student.uni-ulm.de>
|	Added option: If there is new mail, the names of the mailboxes
|	containing new mail can be displayed in a tooltip.
|  3/16/2000 Patch from Stefan Schlott <stefan.schlott@student.uni-ulm.de>
|	Added a check to avoid double counting old mail with status lines
|	from Netscape and standard "Status" mail readers.  Also added a check
|	to avoid counting mime attachments as messages, see is_multipart_mail()
|  2/24/2000 Patch from Darko Koruga <darko.koruga@hermes.si>
|	Continuous mail animation option as long as there is new mail.
| 12/16/1999 Patch from Patrick Morris <pmorris@starnixsolutions.com>
|	check_maildir() accepts procmail '_' messages.
| 10/30/1999 Patch from Patrick Crosby <xb@dotfiles.com>
|	check_maildir() function.
*/

#include "gkrellm.h"
#include "gkrellm_private_proto.h"

#include <utime.h>
#include <dirent.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include "gkrellm_threads.h"

#if defined(__linux__)
#if defined(__GLIBC__) && ((__GLIBC__>2) || (__GLIBC__==2 && __GLIBC_MINOR__>=1))
#define HAVE_GETADDRINFO	1
#endif
#endif

#if defined(__FreeBSD__) && __FreeBSD_version >= 400000
#define HAVE_GETADDRINFO	1
#endif

#if defined(__NetBSD__) || defined(__OpenBSD__)
#define HAVE_GETADDRINFO	1
#endif

#define MUTE_FLAG	-1

/* msg_count_mode has 3 states
*/
#define	MSG_NEW_TOTAL_COUNT	0
#define	MSG_NEW_COUNT		1
#define	MSG_NO_COUNT		2

/* animation_select_mode states
*/
#define	ANIMATION_NONE			0
#define	ANIMATION_ENVELOPE		1
#define	ANIMATION_PENGUIN		2
#define	ANIMATION_BOTH			3

/* # of seconds to wait for a response from a POP3 or IMAP server
*/
#define TCP_TIMEOUT  30


  /* A mailbox type has bits encoding how to check the mailbox (inline code
  |  check or threaded check).
  |  Threaded checks and the fetch program are remote checks
  */
#define	MBOX_CHECK_FETCH		0x1000
#define	MBOX_CHECK_INLINE		0x2000
#define	MBOX_CHECK_THREADED		0x4000
#define MBOX_CHECK_TYPE_MASK	0xf000

  /* Counts for mailboxes created and checked in other plugins can be shown */
#define	MBOX_EXTERNAL_FLAG		0x10

  /* Mailboxes internally checked and created via the Mail->Mailboxes config */
#define MBOX_INTERNAL_FLAG		0x20


  /* Here's the list of all the mailbox types the Mail monitor knows about.
  |  The MBOX_FETCH is a pseudo internal mailbox where the counts read from
  |  the fetch program are kept.  Additionally MBOX_FETCH_TOOLTIP types
  |  are constructed just so the fetch programs output lines can be 
  |  reported in a tooltip.  Real mailboxes that GKrellM creates in its
  |  config and knows how to check have the MBOX_INTERNAL_FLAG set.  And
  |  finally there can be external (plugin) mailboxes created which
  |  can have their check function called at the update intervals.  If the
  |  plugin reports back the count results, the animation/sound can be
  |  triggered for the plugin.  (Don't know if EXTERNAL guys will ever be used)
  */
#define	MBOX_FETCH		(MBOX_CHECK_FETCH)
#define	MBOX_MAILBOX	(MBOX_CHECK_INLINE   | MBOX_INTERNAL_FLAG | 1)
#define	MBOX_MAILDIR	(MBOX_CHECK_INLINE   | MBOX_INTERNAL_FLAG | 2)
#define	MBOX_MH_DIR		(MBOX_CHECK_INLINE   | MBOX_INTERNAL_FLAG | 3)
#define	MBOX_POP3		(MBOX_CHECK_THREADED | MBOX_INTERNAL_FLAG | 4)
#define	MBOX_IMAP		(MBOX_CHECK_THREADED | MBOX_INTERNAL_FLAG | 5)
#define	MBOX_FETCH_TOOLTIP	(6)

#define	MBOX_EXTERNAL_INLINE	(MBOX_CHECK_INLINE   | MBOX_EXTERNAL_FLAG)
#define	MBOX_EXTERNAL_THREADED	(MBOX_CHECK_THREADED | MBOX_EXTERNAL_FLAG)


typedef struct
	{
	gchar	*path;			/* Path to local mailbox	*/
	gchar	*username;		/* Account username			*/
	gchar	*password;		/* Account password			*/
	gchar	*server;		/* Mail Server (POP3/IMAP)	*/
	gchar	*folder;		/* folder if server is IMAP*/
	gint	busy;			/* TRUE if remote mailbox is being checked */
	gint	type;			/* Type of mailbox: MBOX, POP3, IMAP, etc */
	GString	*tcp_in;		/* For reading input from remote mail servers */
	gboolean (*check_func)(); /* Mail check function to call. */
	gpointer data;			/* For external mailboxes (in plugins) */

	gint	mail_count;		/* Total number of messages in a mail box */
	gint	new_mail_count;
	gint	old_mail_count;
	gint	prev_mail_count,
			prev_new_mail_count;
	time_t	last_mtime;		/* File modification time on last check */
	off_t	last_size;		/* File size on last check			*/
	gint	is_internal;	/* Internal mail message (ie: localmachine) */
	gint	need_animation,
			prev_need_animation;
	void	*private;
	}
	Mailbox;

static GList	*mailbox_list;

typedef struct
	{
	gchar	*command;
	GString	*read_gstring;			/* Bytes read from pipe stored here */
	FILE	*pipe;
	}
	Mailproc;

Mailbox			*mail_fetch;		/* Internal mailbox: fetch command */

static Mailproc	mail_user_agent;
static gchar	*mail_notify;		/* Sound		*/

static GdkPixmap	*decal_mail_pixmap;
static GdkBitmap	*decal_mail_mask;

static gint		run_animation,
				decal_frame;

static gint		mail_fetch_timeout = 5;			/* Minutes */
static gint		mail_check_timeout = 4;			/* Seconds */

static Panel	*mail;

static GtkTooltips	*tooltip;

static DecalButton	*mua_button;

static gboolean	mail_visible = FALSE;
static gboolean	enable_mail,
				mute_mode,
				super_mute_mode,
				cont_animation_mode,
				mua_inhibit_mode,		/* Inhibit checking if MUA launched */
				msg_count_mode,
				fetch_check_only_mode,
				reset_remote_mode;

static gboolean	mh_seq_ignore,
				checking_mh_mail;

static gint		animation_select_mode	= ANIMATION_BOTH;

static gboolean	force_mail_check;
static gint		new_mail_count, total_mail_count;
static gint		check_timeout;
static gint		show_tooltip = FALSE;

static gint		anim_frame,
				anim_dir,
				anim_pause;

static gint		style_id;

static void
free_mailbox(Mailbox *mbox)
	{
	/* If user changes mailbox config list while a mailbox thread is busy,
	|  freeing the mbox can cause a segfault.   Probably a rare enough
	|  event that I can just suffer the small memory leak.
	*/
	if (mbox->busy)
		{
		printf("freeing busy mbox\n");
		return;
		}
	if (mbox->path)
		g_free(mbox->path);
	if (mbox->server)
		g_free(mbox->server);
	if (mbox->username)
		g_free(mbox->username);
	if (mbox->password)
		g_free(mbox->password);
	if (mbox->folder)
		g_free(mbox->folder);
	g_free(mbox);
	}

static gboolean
format_remote_mbox_name(Mailbox *mbox, gchar *buf, size_t len)
	{
	if (mbox->folder)
		snprintf(buf, len, "%s-%s@%s", mbox->username,
			mbox->folder, mbox->server);
	else if (mbox->server)
		snprintf(buf, len, "%s@%s", mbox->username, mbox->server);
	else if (mbox->username)
		snprintf(buf, len, "%s", mbox->username);
	else
		{
		snprintf(buf, len, "??");
		return FALSE;
		}
	return TRUE;
	}

  /* Make tooltip visible/invisible and fill it with mailbox names
  |  containing new mail.
  */
static void
update_tooltip()
	{
	GList	*list;
	Mailbox	*mbox;
	GString	*mboxes = NULL;
	gchar	buf[128];
   
	if (show_tooltip)
		{
		mboxes = g_string_sized_new(512);
		for (list = mailbox_list; list; list = list->next)
			{
			mbox = (Mailbox *) list->data;
			if (mbox->new_mail_count > 0)
				{
				if ((    mbox->type == MBOX_MAILBOX
					  || mbox->type == MBOX_MAILDIR
					  || mbox->type == MBOX_MH_DIR
					  || mbox->type == MBOX_EXTERNAL_INLINE
					  || mbox->type == MBOX_EXTERNAL_THREADED
					) && mbox->path
				   )
					snprintf(buf, sizeof(buf), "%s", mbox->path);
				else if (! format_remote_mbox_name(mbox, buf, sizeof(buf)))
					continue;	/* Can't get a name, so no tooltip for you! */

				if (mboxes->len > 0) 
					g_string_append_c(mboxes, '\n');
				g_string_append(mboxes, buf);
							
				if (msg_count_mode == MSG_NEW_TOTAL_COUNT)
					snprintf(buf, sizeof(buf), "(%d/%d)",
						mbox->new_mail_count, mbox->mail_count);
				else
					snprintf(buf, sizeof(buf), "(%d)", mbox->new_mail_count);
				g_string_append(mboxes, buf);
 				}
			}
		}
	if (show_tooltip && mboxes && mboxes->len > 0)
		{
		gtk_tooltips_set_tip(tooltip, mail->drawing_area, mboxes->str, "");
		gtk_tooltips_enable(tooltip);
		}
	else
		gtk_tooltips_disable(tooltip);
	if (mboxes)
		g_string_free(mboxes, TRUE);
	}
	
	
  /* Look at a From line to see if it is valid, lines look like:
  |  From sending_address dayofweek month dayofmonth timeofday year
  |  eg: From bill@gkrellm.net Fri Oct 22 13:52:49 2010
  */
static gint
is_From_line(Mailbox *mbox, gchar *buf)
	{
	gchar	sender[512];
	gint	dayofmonth = 0;

	if (strncmp(buf, "From ", 5))
		return FALSE;

	/* In case sending address missing, look for a day of month
	|  number in field 3 or 4 (0 based).
	*/
	sender[0] = '\0';
	if (sscanf(buf, "%*s %*s %*s %d", &dayofmonth) != 1)
		{
		if (sscanf(buf, "%*s %511s %*s %*s %d", sender, &dayofmonth) != 2)
			return FALSE;
		}
	if (dayofmonth < 1 || dayofmonth > 31)
		return FALSE;
	if (strcmp(sender, "MAILER-DAEMON") == 0)
		mbox->is_internal = TRUE;
	return TRUE;
	}


  /* Routine from Stefan Schlott,
  | Check if this is a Content-Type-line. If it contains a boundary
  | field, copy boundary string to buffer (including two leading and
  | trailing dashes marking the end of a multipart mail) and return
  | true. Otherwise, return false.
  */
static gint
is_multipart_mail(gchar *buf, gchar *separator)
	{
	gchar *fieldstart;
	gchar *sepstart;
	gint  seplen;
	
	if (strncmp(buf, "Content-Type: ", 14) != 0)
		return FALSE;
	if (strncmp(&buf[14], "multipart/", 10) != 0)
		return FALSE;
	fieldstart = &buf[14];
	while (*fieldstart!=0)
		{
		while (*fieldstart!=0 && *fieldstart!=';')
			fieldstart++;
		if (*fieldstart==';') fieldstart++;
		while (*fieldstart!=0 && *fieldstart==' ')
			fieldstart++;
		if (strncmp(fieldstart, "boundary=", 9) == 0)
			{
			sepstart = fieldstart + 9;
			if (sepstart[0]=='"')
				{
				sepstart++;
				seplen = 0;
				while (sepstart[seplen]!='"' && sepstart[seplen]>=32)
					seplen++;
				} else
				{
				seplen = 0;
				while (sepstart[seplen]!=';' && sepstart[seplen]>32)
					seplen++;
				}
			strcpy(separator,"--");
			strncpy(&separator[2],sepstart,seplen);
			strcpy(&separator[seplen+2],"--");
			return TRUE;
			}
		}
	return FALSE;
	}

  /* Hide a password that is embedded in a string.
  */
static void
hide_password(Mailbox *mbox, gchar *line, gint offset)
	{
	gint    n;

	n = strlen(mbox->password);
	while (n--)
		line[offset + n] = '*';
	}

static gint
read_select(gint fd, gchar *buf, size_t size, time_t timeout)
	{ 
	fd_set			readfds;
	struct timeval	tv;
	gint			n	= 0;
         
	FD_ZERO(&readfds); 
	FD_SET(fd, &readfds); 
	tv.tv_sec = timeout; 
	tv.tv_usec = 0; 

	if (select(fd+1, &readfds, NULL, NULL, &tv) > 0) 
		n = read(fd, buf, size);
	return n; 
	}


#if defined (GKRELLM_THREADS)

  /* Read \r\n terminated lines from a remote IMAP or POP3 email server,
  |  and verify the beginning of the line matches a string.
  */
static gboolean
tcp_getline(Mailbox *mbox,  gint fd, gchar *match)
	{
	gchar	buf[256];
	gint	n;
	gchar	c;

	if (mbox->tcp_in)
		mbox->tcp_in = g_string_truncate(mbox->tcp_in, 0);
	else
		mbox->tcp_in = g_string_new("");
	while ((n = read_select(fd, buf, sizeof(buf) - 1, TCP_TIMEOUT)) > 0)
		{
		buf[n] = '\0';
		g_string_append(mbox->tcp_in, buf);
		c = mbox->tcp_in->str[mbox->tcp_in->len - 1];
		if (c == '\r' || c == '\n')
			break;
		}
	if (GK.debug_level & DEBUG_MAIL)
		{
		if (n < 0)
			perror("tcp_getline: ");
		format_remote_mbox_name(mbox, buf, sizeof(buf));
		printf("tcp_getline( %s )<%d>:%s\n", buf,
			mbox->tcp_in->len, mbox->tcp_in->str);
		}
	return (!strncmp(match, mbox->tcp_in->str, strlen(match)) ? TRUE : FALSE);
	}

static void
tcp_sendline(Mailbox *mbox, gint fd, gchar *line)
	{
	gchar	buf[128];

	write(fd, line, strlen (line));
	if (GK.debug_level & DEBUG_MAIL)
		{
		format_remote_mbox_name(mbox, buf, sizeof(buf));
		printf("tcp_sendline( %s ):%s", buf, line);
		}
	}

static gchar	*tcp_error_message[]	=
	{
	N_("Unable to connect."),
	N_("Bad response after connect."),
	N_("Bad response after username."),
	N_("Bad response after password."),
	N_("Bad response after STAT or STATUS."),
	N_("Bad response after UIDL."),
	};

static gboolean
tcp_shutdown(gint fd, Mailbox *mbox, gchar *message)
	{
	gchar	buf[128];

  	if (GK.debug_level & DEBUG_MAIL)
		{
		format_remote_mbox_name(mbox, buf, sizeof(buf));
		printf(_("Mail TCP Error: %s - %s\n"), buf, _(message));
		}
	if (fd != -1)
		close (fd);
	return FALSE;
	}


  /* Added by Michael Bell <mikeb@map.com> 7/4/2000
  |  check_mail_pop3 checks for mail on a pop3 server
  |  mbox->username should be the username on the pop3 server
  |  mbox->password should be the password on the pop3 server
  |  mbox->server should be the fully qualified address of the pop3 server
  */
static gboolean
check_mail_pop3(Mailbox *mbox)
	{
	gint	fd	= -1;
	gchar	line[128];
#ifdef HAVE_GETADDRINFO
	gint 	rv	= 0;
	struct addrinfo	hints, *res, *res0;
#else
	struct hostent	*addr;
	struct sockaddr_in s;
#endif

#ifdef HAVE_GETADDRINFO
	memset(&hints, 0, sizeof(hints));
	hints.ai_family = PF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_protocol = 0;
	if ((rv = getaddrinfo(mbox->server, "110", &hints, &res0)) != 0)
		return tcp_shutdown(fd, mbox, tcp_error_message[0]);

	for (res = res0; res; res = res->ai_next)
		{
		if ((fd = socket(res->ai_family, res->ai_socktype,
				res->ai_protocol)) < 0)
			continue;
		if (connect(fd, res->ai_addr, res->ai_addrlen) >= 0)
			break;
		close(fd);
		fd = -1;
		}
	freeaddrinfo(res0);
	if (fd == -1)
		return tcp_shutdown(fd, mbox, tcp_error_message[0]);
#else
	addr = gethostbyname (mbox->server);
	if (addr)
		memcpy (&s.sin_addr, addr->h_addr, sizeof (struct in_addr));  
	else 
		return tcp_shutdown(fd, mbox, tcp_error_message[0]);

	fd = socket (AF_INET, SOCK_STREAM, 0);
	if (fd < 0) 
		return FALSE;

	s.sin_family = AF_INET;
	s.sin_port = htons (110);

	if (connect (fd, (struct sockaddr *)&s, sizeof (s)) < 0)
		return tcp_shutdown(fd, mbox, tcp_error_message[0]);
#endif

	/* Is the machine we are connected to really a POP3 server?
	*/
	if (! tcp_getline(mbox, fd, "+OK"))
		return tcp_shutdown(fd, mbox, tcp_error_message[1]);

	snprintf (line, sizeof (line), "USER %s\r\n", mbox->username);
	tcp_sendline(mbox, fd, line);
	if (! tcp_getline(mbox, fd, "+OK"))
		return tcp_shutdown(fd, mbox, tcp_error_message[2]);

	snprintf (line, sizeof (line), "PASS %s\r\n", mbox->password);
	write(fd, line, strlen (line));
	if (GK.debug_level & DEBUG_MAIL)
		{
		gchar	buf[128];

		hide_password(mbox, line, 5);
		format_remote_mbox_name(mbox, buf, sizeof(buf));
		printf("tcp_sendline( %s ):%s", buf, line);
		}
	if (! tcp_getline(mbox, fd, "+OK"))
		return tcp_shutdown(fd, mbox, tcp_error_message[3]);

	tcp_sendline(mbox, fd, "STAT\r\n");
	if (! tcp_getline(mbox, fd, "+OK"))
		return tcp_shutdown(fd, mbox, tcp_error_message[4]);

	sscanf(mbox->tcp_in->str, "+OK %d", &mbox->mail_count);
   	snprintf (line, sizeof (line), "UIDL %d\r\n", mbox->mail_count);
	tcp_sendline(mbox, fd, line);
	if (! tcp_getline(mbox, fd, "+OK"))
		return tcp_shutdown(fd, mbox, tcp_error_message[5]);

	tcp_sendline(mbox, fd, "QUIT\r\n");
	close (fd);

	/* Set the new_mail_count only if the UIDL is changed to avoid
	|  re-reporting mail is new after MUA button has been clicked.
	*/
	if (   sscanf(mbox->tcp_in->str, "+OK %*d %128s", line) == 1
		&& gkrellm_dup_string((gchar **) (&mbox->private), line)
	   )
		mbox->new_mail_count = mbox->mail_count;
	return TRUE;
	}
 
  /* Added by Michael Bell <mikeb@map.com> 7/4/2000
  |  check_mail_imap checks for mail on a imap server
  |  mbox->username should be the username on the imap server
  |  mbox->password should be the password on the imap server
  |  mbox->server should be the fully qualified address of the imap server
  |  mbox->folder should be the folder on the imap server
  */
static gboolean
check_mail_imap(Mailbox *mbox)
	{
	gint	fd	= -1;
	gint	messages	= 0;
	gint	unseen		= 0;
	gchar	line[128], *ss;
#ifdef HAVE_GETADDRINFO
	gint	rv	= 0;
	struct addrinfo hints, *res, *res0;
#else
	struct hostent		*addr;
	struct sockaddr_in	s;
#endif

#ifdef HAVE_GETADDRINFO
	memset(&hints, 0, sizeof(hints));
	hints.ai_family = PF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_protocol = 0;
	if ((rv = getaddrinfo(mbox->server, "143", &hints, &res0)) != 0)
		return tcp_shutdown(fd, mbox, tcp_error_message[0]);

	for (res = res0; res; res = res->ai_next)
		{
		if ((fd = socket(res->ai_family, res->ai_socktype,
				res->ai_protocol)) < 0)
			continue;
		if (connect(fd, res->ai_addr, res->ai_addrlen) >= 0)
			break;
		close(fd);
		fd = -1;
		}
	freeaddrinfo(res0);
	if (fd == -1)
		return tcp_shutdown(fd, mbox, tcp_error_message[0]);
#else
	addr = gethostbyname (mbox->server);
	if (addr)
		memcpy (&s.sin_addr, addr->h_addr, sizeof (struct in_addr));  
	else 
		return tcp_shutdown(fd, mbox, tcp_error_message[0]);

	fd = socket (AF_INET, SOCK_STREAM, 0);
	if (fd < 0) 
		return FALSE;

	s.sin_family = AF_INET;
	s.sin_port = htons (143);

	if (connect (fd, (struct sockaddr *)&s, sizeof (s)) < 0)
		return tcp_shutdown(fd, mbox, tcp_error_message[0]);
#endif

	/* Is the machine we are connected to really a IMAP server?
	*/
	if (! tcp_getline(mbox, fd, "* OK"))
		return tcp_shutdown(fd, mbox, tcp_error_message[1]);

	snprintf (line, sizeof (line), ". LOGIN %s %s\r\n", 
				mbox->username, mbox->password);
	write(fd, line, strlen (line));
	if (GK.debug_level & DEBUG_MAIL)
		{
		gchar	buf[128];

		hide_password(mbox, line, strlen(line) - strlen(mbox->password) - 2);
		format_remote_mbox_name(mbox, buf, sizeof(buf));
		printf("tcp_sendline( %s ):%s", buf, line);
		}
	if (! tcp_getline(mbox, fd, ". OK"))
		return tcp_shutdown(fd, mbox, tcp_error_message[2]);

	snprintf (line, sizeof (line), ". STATUS %s (MESSAGES UNSEEN)\r\n",
			mbox->folder);
	tcp_sendline(mbox, fd, line);
	if (! tcp_getline(mbox, fd, "*"))
		return tcp_shutdown(fd, mbox, tcp_error_message[4]);

	ss = strstr(mbox->tcp_in->str, "MESSAGES");
	if (sscanf(ss, "MESSAGES %d", &messages) == 1)
		{
		ss = strstr(mbox->tcp_in->str, "UNSEEN");
		sscanf(ss, "UNSEEN %d", &unseen);
		if (GK.debug_level & DEBUG_MAIL)
			printf(_("Messages: %d\nUnseen: %d\n"), messages, unseen);
		}
	mbox->mail_count = messages;
	mbox->new_mail_count = unseen;

	tcp_sendline(mbox, fd, ". LOGOUT\r\n");
	close (fd);

	return TRUE;
	}
#endif	/* GKRELLM_THREADS */


  /* Check a mh directory for mail. The way that messages are marked as new
  |  depends on the MUA the user is using.  Only .mh_sequences is checked for
  |  now, otherwise all mail found is considered new mail.
  */
static gboolean
check_mh_dir(Mailbox *mbox)
	{
	FILE			*f;
	struct dirent	*entry;
	DIR				*folder;
	gchar			*path, *tok;
	gchar			buf[1024];
	gint			n0, n1;

	mbox->mail_count = mbox->new_mail_count = 0;

	if (!(folder = opendir(mbox->path)))
		return FALSE;
	while ((entry = readdir(folder)) != NULL)
		{
		/* Files starting with a digit are messages. */
		if (isdigit(entry->d_name[0]))
			mbox->mail_count++;
		}
	closedir(folder);
	path = g_strdup_printf("%s/%s", mbox->path, ".mh_sequences");

	/* Some MH dir clients use .mh_sequences, others such as mutt or gnus
	|  do not.  For mixed cases, it's a user option to ignore .mh_sequences.
	*/
	if (!mh_seq_ignore && (f = fopen(path, "r")) != NULL)
		{
		while (fgets(buf, sizeof(buf), f))
			{
			/* Look for unseen sequence like "unseen: 4 7-9 23"
			*/
			if (strncmp(buf, "unseen:", 7))
				continue;
			tok = strtok(buf, " \t\n");
			while ((tok = strtok(NULL, " \t\n")) != NULL)
				{
				if (sscanf(tok, "%d-%d", &n0, &n1) == 2)
					mbox->new_mail_count += n1 - n0 + 1;
				else
					mbox->new_mail_count++;
				}
			break;
			}
		fclose(f);
		}
	else
		mbox->new_mail_count = mbox->mail_count;

	g_free(path);
	return TRUE;
	}


  /* A maildir has new, cur, and tmp subdirectories.  Any file in new
  |  or cur that does not begin with a '.' is a mail message.  It is
  |  suggested that messages begin with the output of time() (9 digits)
  |  but while mutt and qmail use this standard, procmail does not.
  |  maildir(5) says:
  |      It is a good idea for readers to skip all filenames in
  |      new and cur starting with a dot.  Other than this,
  |      readers should not attempt to parse filenames.
  |  So check_maildir() simply looks for files in new and cur.
  |  See also http://cr.yp.to/proto/maildir.html
  */
static gboolean
check_maildir(Mailbox *mbox)
	{
	gchar			path[256];
	struct dirent	*dentry;
	DIR 			*dir;

	mbox->new_mail_count = 0;
	snprintf(path, sizeof(path), "%s/new", mbox->path);
	if ((dir = opendir(path)) != NULL)
		{
		while ((dentry = readdir(dir)) != NULL)
			{
			if (*dentry->d_name != '.' && dentry->d_ino > 0)
				mbox->new_mail_count++;
			}
		closedir(dir);
		}
	mbox->mail_count = mbox->new_mail_count;
	snprintf(path, sizeof(path), "%s/cur", mbox->path);
	if ((dir = opendir(path)) != NULL)
		{
		while ((dentry = readdir(dir)) != NULL)
			{
			if (*dentry->d_name != '.' && dentry->d_ino > 0)
				mbox->mail_count++;
			}
		closedir(dir);
		}

	if (GK.debug_level & DEBUG_MAIL)
		printf(_("mdir %s total=%d old=%d new=%d\n"), mbox->path,
			mbox->mail_count, mbox->old_mail_count, mbox->new_mail_count);
    return TRUE;
	}


  /* Count total mail and old mail in a mailbox.  Old mail can be read
  |  with a Status: R0, or can be accessed and not read with Status: O
  |  So, new mail will be the diff - note that unread mail is not
  |  necessarily new mail.  According to stat() man page:
  |  st_atime is changed by mknod(), utime(), read(), write(), truncate()
  |  st_mtime is changed by mknod(), utime(), write()
  |  But, new mail arriving (writing mailbox) sets st_mtime while reading
  |  the mailbox (mail program reading) sets st_atime.  So the test
  |  st_atime > st_mtime is testing if mbox has been read since last new mail.
  |  Mail readers may restore st_mtime after writting status.
  |  And Netscape mail does status with X-Mozilla-Status: xxxx
  */
static gboolean
status_is_old(gchar *buf)
	{
	if (buf[0] != 'S' && buf[0] != 'X')
		return FALSE;

	/* Standard mail clients
	*/
	if (   strncmp(buf, "Status:", 7) == 0
		&& (strchr(buf, 'R') || strchr(buf, 'O'))	/* R0 0R O */
	   )
		return TRUE;

	/* Netscape
	*/
	if (   strncmp(buf, "X-Mozilla-Status:", 17) == 0
		&& strncmp(buf + 18, "0000", 4) != 0
	   )
		return TRUE;
	return FALSE;
	}

static gboolean
check_mailbox(Mailbox *mbox)
	{
	FILE			*f;
	struct utimbuf	ut;
	struct stat		s;
	gchar			buf[1024];
	gchar			mpart_sep[1024];
	gint			in_header	= FALSE;
	gint			marked_read = FALSE;
	gint			is_multipart = FALSE;

	if (stat(mbox->path, &s) != 0)
		{
		mbox->mail_count = mbox->new_mail_count = 0;
		return FALSE;
		}

	/* Message no-counting mode reports new mail based on mailbox
	|  modified time and size.
	*/
	if (msg_count_mode == MSG_NO_COUNT)
		{
		if (   s.st_size > 0
			&& s.st_size >= mbox->last_size
			&& s.st_mtime >= s.st_atime
		   )
			mbox->new_mail_count = TRUE;
		else
			mbox->new_mail_count = FALSE;
		mbox->mail_count = (s.st_size > 0) ? 1 : 0;	/* boolean, not used */
		mbox->old_mail_count = 0;			/* not used */
		mbox->last_size = s.st_size;
		mbox->last_mtime = s.st_mtime;
		return TRUE;
		}

	/* If the mailboxes have been modified since last check, count
	|  the new/total messages.
	*/
	if (s.st_mtime != mbox->last_mtime || s.st_size != mbox->last_size)
		{
		if ((f = fopen(mbox->path, "r")) == NULL)
			return FALSE;
		mbox->mail_count = 0;
		mbox->old_mail_count = 0;
		while(fgets(buf, sizeof(buf), f))
			{
			if (is_multipart && !in_header)
				{
				/* Skip to last line of multipart mail */
				if (strncmp(buf,mpart_sep,strlen(mpart_sep))==0)
					is_multipart = FALSE;
				}
			else if (buf[0] == '\n')
				{
				in_header = FALSE;
				mbox->is_internal = FALSE;
				}
			else if (is_From_line(mbox, buf))
				{
				mbox->mail_count += 1;
				in_header = TRUE;
				marked_read = FALSE;
				}
			else if (in_header && status_is_old(buf) && !marked_read)
				{
				mbox->old_mail_count += 1;
				marked_read = TRUE;
				}
			else if (in_header && mbox->is_internal)
				{
				if (strncmp(buf, "From: Mail System Internal Data", 31) == 0)
					{
					in_header = FALSE;
					mbox->mail_count -= 1;
					mbox->is_internal = FALSE;
					}
				}
			else if (in_header && is_multipart_mail(buf,mpart_sep))
				{
				is_multipart = TRUE;
				}
			}
		fclose(f);

		/* Restore the mbox stat times for other mail checking programs and
		|  so the (st_atime > st_mtime) animation override below will work.
		*/
		ut.actime = s.st_atime;
		ut.modtime = s.st_mtime;
		utime(mbox->path, &ut);

		mbox->last_mtime = s.st_mtime;
		mbox->last_size = s.st_size;
		if (GK.debug_level & DEBUG_MAIL)
			printf(_("mbox read <%s> total=%d old=%d ==> "),
					mbox->path, mbox->mail_count, mbox->old_mail_count);
		}
	/* Set the animation state when new mail count changes, and override
	|  the animation to false if mbox has been accessed since last modify
	|  (A MUA has probably read the mbox).
	*/
	mbox->new_mail_count = mbox->mail_count - mbox->old_mail_count;
	if (s.st_atime > s.st_mtime)
		{
		mbox->need_animation = FALSE;
		mbox->prev_new_mail_count = mbox->new_mail_count;
		}
	if (GK.debug_level & DEBUG_MAIL)
		printf(_("mbox %s total=%d old=%d new=%d\n"), mbox->path,
			mbox->mail_count, mbox->old_mail_count, mbox->new_mail_count);

    return TRUE;
	}

static Decal	*mail_text_decal;
static Decal	*mail_icon_decal;

static void
draw_mail_text_decal(gint new_mail_count, gint mail_count)
	{
	TextStyle	ts, ts_save;
	gint		x, w;
	Panel		*p;
	Decal		*d;
	Style		*style;
	glong		new_value;
	gchar		buf[32], nbuf[16], tbuf[16];

	p = mail;
	d = mail_text_decal;

	/* Decals have only a single value, but I need a two valued decal here.
	|  So, just pack the two numbers and use that as a test for value.
	*/
	new_value = (mail_count << 16) | new_mail_count;
	if (new_value == d->value)
		return;

	if (GK.debug_level & DEBUG_MAIL)
		printf(_("Draw mail: new=%d count=%d "), new_mail_count, mail_count);

	ts_save = d->text_style;
	if (new_mail_count == MUTE_FLAG)
		{
		sprintf(buf, _("mute"));
		ts = *gkrellm_meter_alt_textstyle(style_id);	/* Use the alt color */
		ts.font = d->text_style.font;
		}
	else
		{
		ts = d->text_style;
		if (msg_count_mode == MSG_NO_COUNT)
			buf[0] = '\0';
		else if (msg_count_mode == MSG_NEW_COUNT)
			{
			if (new_mail_count == 0)
				strcpy(buf, "-");
			else
				snprintf(buf, sizeof(buf), "%d", new_mail_count);
			}
		else /* MSG_NEW_TOTAL_COUNT */
			{
			if (new_mail_count == 0)
				strcpy(nbuf, "-");
			else
				snprintf(nbuf, sizeof(nbuf), "%d", new_mail_count);
			if (mail_count == 0)
				strcpy(tbuf, "-");
			else
				snprintf(tbuf, sizeof(tbuf), "%d", mail_count);
			snprintf(buf, sizeof(buf), "%s/%s", nbuf, tbuf); 
			}
		}
	if (GK.debug_level & DEBUG_MAIL)
		printf(_("string=<%s>\n"), buf);
	w = gdk_string_width(ts.font, buf);
	if (w > d->w)
		{
		ts.font = gkrellm_meter_alt_textstyle(style_id)->font;
		w = gdk_string_width(ts.font, buf);
		}
	style = gkrellm_meter_style(style_id);
	x = UC.chart_width * p->label->position / LABEL_MAX;
	x -= style->margin + w / 2;
	if (p->label->position >= 50)
		x -= mail_icon_decal->w;
	if (x > d->w - w)
		x = d->w - w;
	if (x < 0)
		x = 0;
	d->text_style = ts;
	d->x_off = x;
	gkrellm_draw_decal_text(p, d, buf, new_value);
	d->text_style = ts_save;
	}

static void
mbox_set_animation_state(Mailbox *mbox)
	{
	if (mbox->new_mail_count != mbox->prev_new_mail_count)
		mbox->need_animation =
				(mbox->new_mail_count > mbox->prev_new_mail_count);
	}

static void
update_krell_animation_frame()
	{
	/* Run the animation.  Just go back and forth with a pause on
	|  frames 1 and full_scale - 1 (frames 0 and full_scale are cut off).
	|  Frame 0 is blank anyway, and frame full_scale is just not used.
	*/
	if (anim_pause-- <= 0)
		{
		if (anim_frame <= 1)
			anim_dir = 1;
		if (anim_frame >= KRELL(mail)->full_scale - 1)
			anim_dir = -1;
		anim_frame += anim_dir;
		anim_frame %= KRELL(mail)->full_scale;
		if (anim_frame == 1 || anim_frame == KRELL(mail)->full_scale - 1)
			anim_pause = 4;
		}
	}

  /* I popen the mail_user_agent so I can know if it is running and
  |  the mail_fetch so I can process fetchmail/flist/fetcho/...  output.
  |  Reading the pipes need to not block so GKrellM won't freeze.
  |  So, I need a special non-blocking pipe line reading routine.
  */
static void
pipe_command(Mailproc *mp)
	{
	gchar	buf[128];

	if (mp->pipe)	/* Still running?  */
		{
		if (GK.debug_level & DEBUG_MAIL)
			printf(_("mail pipe_command: old command <%s> still running.\n"),
					mp->command);
		return;
		}
	if (mp->command == NULL || *(mp->command) == '\0')
		return;
	snprintf(buf, sizeof(buf), "%s 2>&1", mp->command);
	if (GK.debug_level & DEBUG_MAIL)
		printf(_("mail launching pipe_command <%s>\n"), buf);
	if ((mp->pipe = popen(buf, "r")) == NULL)
		{
		if (GK.debug_level & DEBUG_MAIL)
			printf(_("mail could not run pipe_command <%s>\n"), buf);
		return;
		}
	if (mp->read_gstring)
		mp->read_gstring = g_string_truncate(mp->read_gstring, 0);
	else
		mp->read_gstring = g_string_new("");
	fcntl(fileno(mp->pipe), F_SETFL, O_NONBLOCK);
	}


  /* Accumulate non-blocking reads from a pipe into a gstring.  Then
  |  try to read a single line from the gstring.  If a line is read,
  |  return the count of chars in the line.  But if no line can be read
  |  and the other end of the pipe has exited (pipe eof), return -1.
  |  Otherwise return 0 to indicate no line is available but the pipe
  |  producer is still alive.
  |  Why this trouble?: the stdio fgets() reading from a non-blocking
  |  stream is not guaranteed to return complete '\n' delimited lines.
  */
static gint
fgets_pipe(gchar *line, gint len, Mailproc *mp)
	{
	gchar	buf[512];
	gint	n;

	if (mp->pipe)
		{
		if (feof(mp->pipe))
			{
			pclose(mp->pipe);
			mp->pipe = NULL;
			}
		else
			{
			n = fread(buf, 1, sizeof(buf) - 1, mp->pipe);
			if (n > 0)
				{
				buf[n] = '\0';
				mp->read_gstring = g_string_append(mp->read_gstring, buf);
				}
			}
		}
	if (getline_from_gstring(&mp->read_gstring, line, len))
		return strlen(line);
	if (!mp->pipe)
		{
		if (mp->read_gstring)
			g_string_free(mp->read_gstring, TRUE);
		mp->read_gstring = NULL;
		return -1;
		}
	return 0;
	}


static gboolean
mua_is_launched()
	{
	gchar	buf[128];
	gint	n;

	if (mail_user_agent.pipe == NULL)
		return FALSE;
	while ((n = fgets_pipe(buf, sizeof(buf), &mail_user_agent)) > 0)
		;
	return (n < 0) ? FALSE : TRUE;
	}

  /* Read mail_fetch pipe and if fetch_check_only_mode, look for output
  |  of fetchmail -c or flist so I can report mail from these programs.
  |
  | eg. for MH mail, the nmh program flist -all (from the man page):
  |   /work/Mail  has  5 in sequence unseen (private); out of  46
  |   inbox+      has 10 in sequence unseen          ; out of 153
  |   junklist    has  0 in sequence unseen          ; out of  63
  |
  |  For fetchmail, if no remote mail, I get:
  |		fetchmail: No mail for billw at mail.wt.net
  |  If remote mail i will get lines like:
  |		1 message for billw at mail.wt.net (32743 octets).
  |	  or, as reported by a user, there could be:
  |		26 messages (25 seen) for billw at mail.wt.net
  |  If the remote mail is fetched, I get additional lines like:
  |		reading message 1 of 1 (32743 octets) .................... flushed
  |  Note: above 26 messages (25 seen) should show as 1/26 on panel.
  |
  |  And fetchmail can't make up its mind.  A user gets this with 5.2.0:
  |  fetchmail: 5.2.0 querying xx.yy.net (protocol POP3) at Thu, 20 Jan 2000...
  |  fetchmail: 2 messages for uuu at xx.yy.net (20509 octets).
  */

  /* Since there is now internal POP3 and IMAP checking, this will not be
  |  used as much, but there is still flist, using fetchmail for ssl, etc.
  |  Anyway, here's the code to grok the above mess!
  */
static gboolean
parse_fetch_line(gchar *line, gint *msg, gint *seen)
	{
	gchar	*s, *p, *buf;
	gint	n, n1, n2;
	gint	tok_count	= 0;
	gint	state		= 0,
			seen_flag	= FALSE,
			unseen_flag	= FALSE;

	*msg = 0;
	*seen = 0;
	n1 = n2 = 0;
	buf = g_strdup(line);
	s = strtok(buf, " \t()\n");

	/* Trap out badly set Fetch/check option
	*/
	if (!s || !strcmp(s, "reading") || !strcmp(s, "skipping"))
		{
		g_free(buf);
		return FALSE;
		}

	if (GK.debug_level & DEBUG_MAIL)
		printf(_("  parse["));
	while (s)
		{
		if (++tok_count > 3 && state == 0)	/* need a int within 3 tokens */
			break;
		n = strtol(s, &p, 0);
		if (*p == ' ' || *p == '\t' || *p == '\0' || *p == '\n')
			{		/* Have an integer, and not a x.y version number */
			if (GK.debug_level & DEBUG_MAIL)
				printf(_("*<%s>,st=%d,"), s, state);
			if (state == 0)
				n1 = n;
			else if (state == 1)
				n2 = n;
			if (GK.debug_level & DEBUG_MAIL)
				printf(_("n1=%d,n2=%d state=%d*"), n1, n2, state);
			++state;
			}
		else if (strcmp(s, "seen") == 0)
			seen_flag = TRUE;
		else if (strcmp(s, "unseen") == 0)
			unseen_flag = TRUE;
		s = strtok(NULL, " \t()\n");
		}
	if (state > 1 && seen_flag)	/* 26 messages (25 seen) ... */
		{
		*msg = n1;
		*seen = n2;
		}
	else if (state > 1 && unseen_flag)	/* /xxx has 5 in sequence unseen ... */
		{
		*msg = n2;
		*seen = n2 - n1;
		}
	else if (state > 0)			/*     1 message for billw at ... */
		*msg = n1;				/* or  Fetchmail: 1 message for .... */
	if (GK.debug_level & DEBUG_MAIL)
		printf(_("]snf=%d sunf=%d msg=%d seen=%d STATE=%d\n"),
				seen_flag, unseen_flag, *msg, *seen, state);
	g_free(buf);
	return TRUE;
	}


static gint
compare_mailboxes(gconstpointer a, gconstpointer b)
	{
	gchar a_name[128];
	gchar b_name[128];

	format_remote_mbox_name((Mailbox *)a, a_name, sizeof(a_name));
	format_remote_mbox_name((Mailbox *)b, b_name, sizeof(b_name));

	return strcmp(a_name, b_name);
	}

static void
make_fetch_tooltip(gchar *line, gint msg, gint seen)
	{
	Mailbox		*mbox;
	GList		*old_mbox_pointer;
	gchar		buf[64], *s;

	mbox = g_new0(Mailbox, 1);
	mbox->type = MBOX_FETCH_TOOLTIP;

	if ((s = strstr(line, "sequence unseen")) != NULL)	/* flist */
		{
		sscanf(line, "%64s", buf);
		mbox->username = g_strdup(buf);
		}
	else if ((s = strstr(line, " for ")) != NULL)		/* fetchmail */
		{
		sscanf(s + 5, "%64s", buf);
		mbox->username = g_strdup(buf);
		if ((s = strstr(line, " at ")) != NULL)
			{
			sscanf(s + 4, "%64s", buf);
			mbox->server = g_strdup(buf);
			}
		if ((s = strstr(line, "(folder ")) != NULL)
			{
			sscanf(s + 8, "%64[^)]", buf);
			mbox->folder = g_strdup(buf);
			}
		}
	else
		{
		free_mailbox(mbox);
		return;
		}
	old_mbox_pointer = g_list_find_custom(mailbox_list, mbox,
				(GCompareFunc) compare_mailboxes);
	if (old_mbox_pointer)
		{
		free_mailbox(mbox);
		mbox = (Mailbox *) old_mbox_pointer->data;
		if (mbox->type == MBOX_FETCH_TOOLTIP)
			{
			if (GK.debug_level & DEBUG_MAIL)
				printf(_("fetch tooltip - found old mailbox: %s, %s, %s\n"),
						mbox->folder, mbox->username, mbox->server);
			mbox->mail_count = msg;
			mbox->new_mail_count = msg - seen;
			}
		}
	else
		{
		if (GK.debug_level & DEBUG_MAIL)
			printf(_("fetch tooltip - inserting new mailbox: %s, %s, %s\n"),
						mbox->folder, mbox->username, mbox->server);
		mbox->mail_count = msg;
		mbox->new_mail_count = msg - seen;
		g_list_insert_sorted(mailbox_list, mbox, (GCompareFunc)(compare_mailboxes));
		}
	}

  /* Read output lines from the fetch/check program.  If fetch_check_only_mode
  |  is set, parse the lines to try to read message counts so they can be
  |  reported ("fetchmail -c" or equiv should be configured).  However, if
  |  the mode is not set, then just read lines and waste them because mail
  |  is presumably being downloaded into local mailboxes.
  */
static void
read_mail_fetch()
	{
	Mailproc	*mp		= (Mailproc *) mail_fetch->private;
	gchar		buf[128];
	gint		n, msg, seen;

	while ((n = fgets_pipe(buf, sizeof(buf), mp)) > 0)	/* non-blocking */
		{
		if (GK.debug_level & DEBUG_MAIL)
			printf(_("read_mail_fetch(%d): %s\n"), fetch_check_only_mode, buf);
		if (fetch_check_only_mode)
			{
			if (parse_fetch_line(buf, &msg, &seen))
				make_fetch_tooltip(buf, msg, seen);
			if (msg > 0)
				{
				mail_fetch->mail_count += msg;
				mail_fetch->old_mail_count += seen;
				if (GK.debug_level & DEBUG_MAIL)
					printf(_("read_mail_fetch: fetch_messages=%d fetch_seen=%d\n"),
						mail_fetch->mail_count, mail_fetch->old_mail_count);
				}
			}
		}
	if (n == 0 && (GK.debug_level & DEBUG_MAIL))
		printf(_("read_mail_fetch: no read data.\n"));

	/* When fetch program is done, flag not busy so new counts will be used.
	*/
	if (n < 0)
		{
		if (fetch_check_only_mode)
			{
			mail_fetch->new_mail_count =
					mail_fetch->mail_count - mail_fetch->old_mail_count;
			mbox_set_animation_state(mail_fetch);
			if (GK.debug_level & DEBUG_MAIL)
				printf(_("read_mail_fetch: total=%d old=%d new=%d\n"),
					mail_fetch->mail_count,
					mail_fetch->old_mail_count, mail_fetch->new_mail_count);
			}
		mail_fetch->busy = FALSE;
		}
	}

static void
reset_mail_fetch()
	{
	Mailproc	*mp	= (Mailproc *) mail_fetch->private;

	if (mp->pipe)
		pclose(mp->pipe);
	mp->pipe = NULL;
	if (mp->read_gstring)
		g_string_free(mp->read_gstring, TRUE);
	mp->read_gstring = NULL;
	mail_fetch->mail_count = 0;
	mail_fetch->new_mail_count = 0;
	mail_fetch->old_mail_count = 0;
	mail_fetch->need_animation = FALSE;
	mail_fetch->busy = FALSE;
	force_mail_check = TRUE;
	}

static gboolean
run_fetch_program()
	{
	Mailbox		*mbox;
	Mailproc	*mp	= (Mailproc *) mail_fetch->private;
	GList		*list;

	if (   !mail_fetch->busy && !mp->read_gstring
		&& mp->command && *(mp->command) != '\0'
	   )
		{
		mail_fetch->busy = TRUE;
		for (list = mailbox_list; list; list = list->next)
			{
			mbox = (Mailbox *) list->data;
			if (mbox->type == MBOX_FETCH || mbox->type == MBOX_FETCH_TOOLTIP)
				{
				mbox->mail_count = 0;
				mbox->new_mail_count = 0;
				mbox->old_mail_count = 0;
				}
			}
		pipe_command(mp);
		if (mp->pipe)
			return TRUE;
		}
	return FALSE;
	}

#if defined (GKRELLM_THREADS)
  /* This routine is run as a thread so GKrellM does not have to wait
  |  for the results of a mail checking function.
  */
static void
remote_mail_check_func(void *data)
	{
	Mailbox	*mbox	= (Mailbox *) data;
	gint	external;

	external = mbox->type & MBOX_EXTERNAL_FLAG;
	if ( (*(mbox->check_func))(external ? mbox->data : mbox) == FALSE)
		{
		mbox->mail_count = 0;
		mbox->new_mail_count = 0;
		mbox->need_animation = FALSE;
		}
	else
		mbox_set_animation_state(mbox);
	mbox->busy = FALSE;
	}

static void
run_mail_check_thread(Mailbox *mbox)
	{
	gpointer	thread_id;

	if (!mbox->busy && mbox->check_func)
		{
		mbox->busy = TRUE;
		thread_id = new_thread(remote_mail_check_func, mbox);
		}
	}
#endif

static void
update_mail(void)
	{
	Mailbox		*mbox;
	GList		*list;
	gchar		cmd[256];
	gint		external, prev_new_mail_count, any_busy;
	gint		local_check		= FALSE,
				remote_check	= FALSE;
	static gint	second_count,
				minute_count,
				sound_inhibit;

	if (enable_mail == FALSE || mailbox_list == NULL)
		return;
	if (GK.minute_tick && (++minute_count % mail_fetch_timeout) == 0)
		remote_check = TRUE;
	if (GK.second_tick && (++second_count % mail_check_timeout) == 0)
		local_check = TRUE;
	if (remote_check || local_check)
		mua_is_launched();		/* update pipe, avoid lingering zombie */

	if (   force_mail_check
		|| (   GK.second_tick
			&& !(mute_mode && super_mute_mode)
			&& !(mua_inhibit_mode && mail_user_agent.pipe)
		   )
	   )
		{
		prev_new_mail_count = new_mail_count;
		total_mail_count = 0;
		new_mail_count = 0;
		run_animation = FALSE;
		remote_check |= force_mail_check;
		local_check |= force_mail_check;
		if (force_mail_check)
			minute_count = second_count = 0;
		force_mail_check = FALSE;
		any_busy = FALSE;

		for (list = mailbox_list; list; list = list->next)
			{
			mbox = (Mailbox *) list->data;
			external = mbox->type & MBOX_EXTERNAL_FLAG;
			switch (mbox->type & MBOX_CHECK_TYPE_MASK)
				{
				case MBOX_CHECK_FETCH:
					if (((Mailproc *)(mail_fetch->private))->read_gstring)
						read_mail_fetch();
					if (remote_check)
						(*mbox->check_func)(mbox);
					break;
					
				case MBOX_CHECK_INLINE:	/* Local mailbox or maildir check */
					if (local_check)
						{
						if (mbox->check_func)
							(*mbox->check_func)(external ? mbox->data : mbox);
						mbox_set_animation_state(mbox);
						}
					break;
#if defined (GKRELLM_THREADS)
				case MBOX_CHECK_THREADED:
					if (remote_check)
						run_mail_check_thread(mbox);
					break;
#endif
				default:	/* Unknown or pseudo mail box type */
					continue;
				}
			/* If a mailbox check is busy (thread or remote fetch program
			|  running), use previous counts until the check is done.
			*/
			if (mbox->busy)
				{
				total_mail_count += mbox->prev_mail_count;
				new_mail_count += mbox->prev_new_mail_count;
				run_animation |= mbox->prev_need_animation;
				any_busy = TRUE;
				}
			else
				{
				total_mail_count += mbox->mail_count;
				new_mail_count += mbox->new_mail_count;
				if (mbox->new_mail_count && cont_animation_mode)
					mbox->need_animation = TRUE;
				run_animation |= mbox->need_animation;
				mbox->prev_mail_count = mbox->mail_count;
				mbox->prev_new_mail_count = mbox->new_mail_count;
				mbox->prev_need_animation = mbox->need_animation;
				}
			}
		if ((GK.debug_level & DEBUG_MAIL) && (local_check || remote_check))
			printf(_("**Mail check totals: total=%d new=%d anim=%d\n"),
				total_mail_count, new_mail_count, run_animation);

		if (   prev_new_mail_count != new_mail_count
			|| tooltip->active_tips_data == NULL
		   )
			update_tooltip();

		/* Run the notify (sound) command if the new mail count goes up,
		|  and if the various modes permit.  Ensure a sound command is
		|  issued only once per remote check interval by locking it out
		|  if a new mail count triggers the sound and there are other
		|  remote threads still busy.
		*/
		if (   !sound_inhibit && !mute_mode
			&& new_mail_count > prev_new_mail_count
			&& mail_notify && mail_notify[0] != '\0'
		   )
			{
			snprintf(cmd, sizeof(cmd), "%s &", mail_notify);
			system(cmd);
			if (any_busy)
				sound_inhibit = TRUE;
			}
		if (!any_busy)
			sound_inhibit = FALSE;
		}

	if (mute_mode && (GK.timer_ticks % 15) < 3)	/* Asymmetric blink */
		draw_mail_text_decal(MUTE_FLAG, MUTE_FLAG);
	else
		draw_mail_text_decal(new_mail_count, total_mail_count);

	if (run_animation && (animation_select_mode & ANIMATION_PENGUIN))
		update_krell_animation_frame();
	else
		anim_frame = 0;

	if ((GK.timer_ticks % GK.decal_mail_delay) == 0)
		{
		if (run_animation)
			{
			if (animation_select_mode & ANIMATION_ENVELOPE)
				{
				++decal_frame;
				if (decal_frame >= GK.decal_mail_frames)
					decal_frame = 1;
				}
			else
				decal_frame = 0;
			}
		else
			decal_frame = 1;
		}
	gkrellm_draw_decal_pixmap(mail, DECAL(mail), decal_frame);

	/* All the animation frame drawing is done with the general krell code.
	*/
	KRELL(mail)->previous = 0;
	gkrellm_update_krell(mail, KRELL(mail), anim_frame);
	gkrellm_draw_layers(mail);
	}


static gint
mail_expose_event(GtkWidget *widget, GdkEventExpose *ev)
	{
	if (widget == mail->drawing_area)
		gdk_draw_pixmap(widget->window, GK.draw1_GC, mail->pixmap,
			ev->area.x, ev->area.y, ev->area.x, ev->area.y,
			ev->area.width, ev->area.height);
	return FALSE;
	}

  /* The mua launch button also optionally stops animations and resets
  |  remote mail counts.  So this routine decides if it should be enabled.
  */
static void
set_mua_button_sensitivity()
	{
	if (   *(mail_user_agent.command) || reset_remote_mode
		|| (animation_select_mode && !cont_animation_mode)
	   )
		mua_button->sensitive = TRUE;
	else
		mua_button->sensitive = FALSE;
	}


  /* Callback for the message count decal button.
  */
static void
cb_mail_button(DecalButton *button)
	{
	GList	*list;
	Mailbox	*mbox;

	if (reset_remote_mode)
		{
		for (list = mailbox_list; list; list = list->next)
			{
			mbox = (Mailbox *) list->data;
			if (   (mbox->type & MBOX_CHECK_THREADED)
				|| (mbox->type & MBOX_CHECK_FETCH)
				|| mbox->type == MBOX_FETCH_TOOLTIP		/* A pseudo mbox */
			   )
				{
				GDK_THREADS_ENTER();
				mbox->mail_count = 0;
				mbox->new_mail_count = 0;
				mbox->old_mail_count = 0;
				mbox->need_animation = FALSE;
				GDK_THREADS_LEAVE();
				}
			}
		}
	if (!cont_animation_mode)
		{
		for (list = mailbox_list; list; list = list->next)
			{
			mbox = (Mailbox *) list->data;
			mbox->need_animation = FALSE;
			}
		run_animation = FALSE;
		}
	if (mua_is_launched() == FALSE)
		pipe_command(&mail_user_agent);
	else
		{
		check_timeout = 0;
		if (GK.debug_level & DEBUG_MAIL)
			printf(_("Mail user agent is already running.\n"));
		}
	}

  /* Callback for any button clicks on the Mail panel.  Must exclude
  |  button 1 clicks on the message decal button.
  */
static gint
cb_panel_press(GtkWidget *widget, GdkEventButton *ev)
	{
	Decal	*d 	= mail_icon_decal;

	/* button 2 or 3 anywhere in the panel toggles the mute mode.
	*/
	if (ev->button == 2 || ev->button == 3)
		{
		mute_mode = 1 - mute_mode;

		/* If coming out of mute mode and mail checking was inhibited,
		|  force a check.
		*/
		if (   ! mute_mode && super_mute_mode
			&& (! mua_is_launched() || ! mua_inhibit_mode)
		   )
			force_mail_check = TRUE;
		}
	/* Button 1 press on the envelope - must exclude the message count button.
	*/
	if (ev->button == 1 && ev->x >= d->x && ev->x < d->x + d->w)
		force_mail_check = TRUE;
	return TRUE;
	}	

static Mailbox *
add_mailbox(gchar *arg)
	{
	Mailbox		*mbox;
	gchar		parms[4][CFG_BUFSIZE];
	gint		n;

	mbox = g_new0(Mailbox, 1);
	n = sscanf(arg, "%s %s %s %s", parms[0], parms[1], parms[2], parms[3]);
	switch (n)
		{
		case 1:
			if (isdir(parms[0], "new"))
				{
				mbox->type = MBOX_MAILDIR;
				mbox->check_func = check_maildir;
				}
			else if (isdir(parms[0], NULL))
				{
				mbox->type = MBOX_MH_DIR;
				mbox->check_func = check_mh_dir;
				checking_mh_mail = TRUE;
				}
			else
				{
				mbox->type = MBOX_MAILBOX;
				mbox->check_func = check_mailbox;
				}
			mbox->path = g_strdup(parms[0]);
			break;

#if defined (GKRELLM_THREADS)
		case 3:
			mbox->type = MBOX_POP3;
			mbox->server = g_strdup(parms[0]);
			mbox->username = g_strdup(parms[1]);
			mbox->password = g_strdup(parms[2]);
			mbox->check_func = check_mail_pop3;
			break;

		case 4:
			mbox->type = MBOX_IMAP;
			mbox->server = g_strdup(parms[0]);
			mbox->folder = g_strdup(parms[1]);
			mbox->username = g_strdup(parms[2]);
			mbox->password = g_strdup(parms[3]);
			mbox->check_func = check_mail_imap;
			break;
#endif

		default:	/* Invalid mailbox line */
			g_free(mbox);
			return NULL;
		}
	mailbox_list = g_list_append(mailbox_list, mbox);
	return mbox;
	}

static void
create_mail(GtkWidget *vbox, gint first_create)
	{
	Style		*style;
	Panel		*p;
	gchar		*s, buf[64];
	gint		x, w;
	static GdkImlibImage	*krell_penguin_image;


	if (first_create)
		mail = gkrellm_panel_new0();
	else
		{
		gkrellm_destroy_decal_list(mail);
		gkrellm_destroy_krell_list(mail);
		}
	p = mail;
	if (! mailbox_list->next)	/* First mailbox is internal fetch */
		{
		if ((s = getenv("MAIL")) == NULL)
			if ((s = getenv("USER")) != NULL)
				{
				if (isdir("/var/mail", NULL))
					snprintf(buf, sizeof(buf), "/var/mail/%s", s);
				else
					snprintf(buf, sizeof(buf), "/var/spool/mail/%s", s);
				s = buf;
				}
		if (s)
			add_mailbox(s);
		}

	style = gkrellm_meter_style(style_id);

	if (krell_penguin_image)
		{
		gdk_imlib_kill_image(krell_penguin_image);
		krell_penguin_image = NULL;
		}
	gkrellm_load_image("krell_penguin", NULL, &krell_penguin_image,
			MAIL_STYLE_NAME);
	gkrellm_create_krell(p,
			krell_penguin_image ? krell_penguin_image
								: gkrellm_krell_meter_image(style_id),
			style);
	KRELL(p)->full_scale = style->krell_depth - 1;

	/* But for 0.9.x theme compatibility, I should #include decal_mail.xpm
	|  and use it here.  But has to be done in config.c for now.
	|  This will grab new theme style net/decal_mail.png images if
	|  present and reload decal_mail_image.
	*/
	gkrellm_load_image("decal_mail", NULL, &GK.decal_mail_image,
			MAIL_STYLE_NAME);
	gkrellm_render_to_pixmap(GK.decal_mail_image, &decal_mail_pixmap,
			&decal_mail_mask, 0, 0);
	mail_icon_decal = gkrellm_create_decal_pixmap(p, decal_mail_pixmap,
				decal_mail_mask, GK.decal_mail_frames, style, -1, -1);

	/* The new/total mail text needs to be a decal because the text changes
	|  and must be drawn as a layer in update_layers().
	|  Calculate x,w override values. Cannot override w after the create.
	*/
	x = style->margin;
	if (style->label_position >= 50)
		x += mail_icon_decal->w;
	w = UC.chart_width - mail_icon_decal->w - 2 * style->margin;

	p->textstyle = gkrellm_meter_textstyle(style_id);

    mail_text_decal = gkrellm_create_decal_text(p, "0", p->textstyle,
                style, x, -1, w);    /* -1 means use y default */

	gkrellm_configure_panel(p, NULL, style);
	gkrellm_create_panel(vbox, p, gkrellm_bg_meter_image(style_id));

	/* Center the decals with respect to each other.
	*/
	if (mail_icon_decal->h > mail_text_decal->h)
		mail_text_decal->y += (mail_icon_decal->h - mail_text_decal->h) / 2;
	else
		mail_icon_decal->y += (mail_text_decal->h - mail_icon_decal->h) / 2;

	mua_button = gkrellm_put_decal_in_meter_button(p, mail_text_decal,
			cb_mail_button, NULL);
	set_mua_button_sensitivity();

	if(enable_mail)
		{
		gkrellm_monitor_height_adjust(p->h);
		mail_visible = TRUE;
		}
	else
		gtk_widget_hide(p->hbox);

	if (first_create)
		{
		gtk_signal_connect(GTK_OBJECT (p->drawing_area), "expose_event",
				(GtkSignalFunc) mail_expose_event, NULL);
		gtk_signal_connect(GTK_OBJECT(p->drawing_area),"button_press_event",
				(GtkSignalFunc) cb_panel_press, NULL);
		tooltip=gtk_tooltips_new();
		}
	}


static void
save_mail_config(FILE *f)
	{
	Mailbox		*mbox;
	Mailproc	*mp		= (Mailproc*) mail_fetch->private;
	GList		*list;

	for (list = mailbox_list; list; list = list->next)
		{
		mbox = (Mailbox *) list->data;

		switch (mbox->type)
			{
			case MBOX_MAILBOX:
			case MBOX_MAILDIR:
			case MBOX_MH_DIR:
				fprintf(f, "mail mailbox %s\n", mbox->path);
				break;
			case MBOX_POP3:
				fprintf(f, "mail mailbox %s %s %s\n", mbox->server,
					mbox->username, mbox->password);
				break;
			case MBOX_IMAP:
				fprintf(f, "mail mailbox %s %s %s %s\n", mbox->server,
					mbox->folder, mbox->username, mbox->password);
				break;
			default:
				break;
			}
		}
	fprintf(f, "mail mua %s\n", mail_user_agent.command);
	fprintf(f, "mail notify %s\n", mail_notify);
	fprintf(f, "mail fetch_command %s\n", mp->command);
	fprintf(f, "mail fetch_timeout %d\n", mail_fetch_timeout);
	fprintf(f, "mail check_timeout %d\n", mail_check_timeout);
	fprintf(f, "mail msg_count_mode %d\n", msg_count_mode);
	fprintf(f, "mail animation_select_mode %d\n", animation_select_mode);
	fprintf(f, "mail fetch_check_only_mode %d\n", fetch_check_only_mode);
	fprintf(f, "mail reset_remote_mode %d\n", reset_remote_mode);
	fprintf(f, "mail enable %d %d %d\n", enable_mail, super_mute_mode,
						mua_inhibit_mode);
	fprintf(f, "mail animation_continuous %d\n", cont_animation_mode);
	fprintf(f, "mail show_tooltip %d\n", show_tooltip);
	fprintf(f, "mail mh_seq_ignore %d\n", mh_seq_ignore);
	}

static void
load_mail_config(gchar *arg)
	{
	Mailbox		*mbox	= NULL;
	Mailproc	*mp		= (Mailproc*) mail_fetch->private;
	gchar		mail_config[32], item[CFG_BUFSIZE];
	gint		n;

	n = sscanf(arg, "%31s %[^\n]", mail_config, item);
	if (n == 2)
		{
		if (strcmp(mail_config, "mailbox") == 0)
			mbox = add_mailbox(item);
		else if (strcmp(mail_config, "mua") == 0)
			mail_user_agent.command = g_strdup(item);
		else if (strcmp(mail_config, "notify") == 0)
			mail_notify = g_strdup(item);
		else if (strcmp(mail_config, "fetch_command") == 0)
			mp->command = g_strdup(item);
		else if (strcmp(mail_config, "fetch_timeout") == 0)
			sscanf(item, "%d", &mail_fetch_timeout);
		else if (strcmp(mail_config, "check_timeout") == 0)
			sscanf(item, "%d", &mail_check_timeout);
		else if (strcmp(mail_config, "msg_count_mode") == 0)
			sscanf(item, "%d", &msg_count_mode);
		else if (strcmp(mail_config, "animation_select_mode") == 0)
			sscanf(item, "%d", &animation_select_mode);
		else if (strcmp(mail_config, "fetch_check_only_mode") == 0)
			sscanf(item, "%d", &fetch_check_only_mode);
		else if (strcmp(mail_config, "reset_remote_mode") == 0)
			sscanf(item, "%d", &reset_remote_mode);
		else if (strcmp(mail_config, "enable") == 0)
			sscanf(item, "%d %d %d", &enable_mail, &super_mute_mode,
						&mua_inhibit_mode);
		else if (strcmp(mail_config, "animation_continuous") == 0)
			sscanf(item, "%d", &cont_animation_mode);
		else if (strcmp(mail_config, "show_tooltip") == 0)
			sscanf(item, "%d", &show_tooltip);
		else if (strcmp(mail_config, "mh_seq_ignore") == 0)
			sscanf(item, "%d", &mh_seq_ignore);
		if (GK.debug_level & DEBUG_MAIL)
			{
			if (mbox && mbox->password)
				hide_password(mbox, item,
						strlen(item) - strlen(mbox->password));
			printf(_("mail_config=%s item=<%s>\n"), mail_config, item);
			}
		}
	}


/* ---------------------------------------------------------------------*/
static GtkWidget		*mbox_path_entry;

static GtkWidget		*mail_user_agent_entry;
static GtkWidget		*mail_fetch_entry;
static GtkWidget		*mail_notify_entry;
static GtkWidget		*enable_mail_button;
static GtkWidget		*enabel_cont_anim_button;
static GtkWidget		*super_mute_button;
static GtkWidget		*mbox_clist;
static GtkWidget		*mail_fetch_spin_button;
static GtkWidget		*mail_check_spin_button;
static GtkWidget		*msg_count_button[3];
static GtkWidget		*animation_select_button[4];
static GtkWidget		*check_only_button;
static GtkWidget		*reset_remote_button;
static GtkWidget		*mua_inhibit_button;
static GtkWidget		*show_tooltip_button;
static GtkWidget		*mh_seq_ignore_button;

static GList			*password_list;

static gint				mailboxes_modified,
						selected_row;

static void
cb_mailbox_selected(GtkWidget *clist, gint row, gint column,
					GdkEventButton *bevent, gpointer data)
	{
	gchar	*s;

	gtk_clist_get_text(GTK_CLIST(clist), row, 0, &s);
	gtk_entry_set_text(GTK_ENTRY(mbox_path_entry), s);
	selected_row = row;
	}

static void
cb_mailbox_unselected(GtkWidget *clist, gint row, gint column,
					GdkEventButton *bevent, gpointer data)
	{
	selected_row = -1;
	gtk_entry_set_text(GTK_ENTRY(mbox_path_entry), "");
	}

static void
cb_enter_mailbox()
	{
	gchar			*buf[2], *s;

	s = gkrellm_entry_get_text(&mbox_path_entry);
	if (*s == '\0')
		return;
	buf[0] = s;
	buf[1] = NULL;
	if (selected_row >= 0)
		{
		gtk_clist_set_text(GTK_CLIST(mbox_clist), selected_row, 0, s);
		gtk_clist_unselect_row(GTK_CLIST(mbox_clist), selected_row, 0);
		selected_row = -1;
		}
	else
		{
		gtk_clist_append(GTK_CLIST(mbox_clist), buf);
		password_list = g_list_append(password_list, NULL);
		}
	gtk_entry_set_text(GTK_ENTRY(mbox_path_entry), "");
	mailboxes_modified = TRUE;
	}


static void
cb_delete_mailbox(GtkWidget *widget, gpointer *data)
	{
	GList	*list;
	gint	row	= selected_row;

	gtk_entry_set_text(GTK_ENTRY(mbox_path_entry), "");
	if (row >= 0)
		{
		gtk_clist_remove(GTK_CLIST(mbox_clist), row);
		mailboxes_modified = TRUE;
		list = g_list_nth(password_list, row);
		selected_row = -1;
		if (list)
			{
			if (list->data)
				g_free((gchar *) list->data);
			password_list = g_list_remove_link(password_list, list);
			g_list_free(list);
			}
		}
	}

#if 0
static void
dump_mailboxes()
	{
	Mailbox	*mbox;
	GList	*list;
	gchar	buf[64];

	for (list = mailbox_list->next; list; list = list->next)
		{
		mbox = (Mailbox *) list->data;
		format_remote_mbox_name(mbox, buf, 64);
		printf("%x path=%s remotename=%s\n", mbox->type, mbox->path, buf);
		}
	printf("\n");
	}
#endif

static void 
apply_mail_config()
	{
	Mailbox			*mbox;
	Mailproc		*mp		= (Mailproc *) mail_fetch->private;
	GList			*list;
	gchar			*s;
	gint			r, i;
	GtkSpinButton	*spin;

	super_mute_mode = GTK_TOGGLE_BUTTON(super_mute_button)->active;
	enable_mail = GTK_TOGGLE_BUTTON(enable_mail_button)->active;
	cont_animation_mode = GTK_TOGGLE_BUTTON(enabel_cont_anim_button)->active;
	gkrellm_enable_visibility(enable_mail, &mail_visible,
					mail->hbox, mail->h);
	for (i = 0; i < 3; ++i)
		if (GTK_TOGGLE_BUTTON(msg_count_button[i])->active)
			msg_count_mode = i;
	for (i = 0; i < 4; ++i)
		if (GTK_TOGGLE_BUTTON(animation_select_button[i])->active)
			animation_select_mode = i;
	mail_text_decal->value += 1000;		/* Force a redraw */
	mua_inhibit_mode = GTK_TOGGLE_BUTTON(mua_inhibit_button)->active;
	fetch_check_only_mode = GTK_TOGGLE_BUTTON(check_only_button)->active;
	reset_remote_mode = GTK_TOGGLE_BUTTON(reset_remote_button)->active;
	show_tooltip = GTK_TOGGLE_BUTTON(show_tooltip_button)->active;
	if (mh_seq_ignore_button)
		mh_seq_ignore = GTK_TOGGLE_BUTTON(mh_seq_ignore_button)->active;
	update_tooltip();

	s = gkrellm_entry_get_text(&mail_user_agent_entry);
	gkrellm_dup_string(&mail_user_agent.command, s);
	set_mua_button_sensitivity();
	
	s = gkrellm_entry_get_text(&mail_fetch_entry);
	if (gkrellm_dup_string(&mp->command, s))
		reset_mail_fetch();

	s = gkrellm_entry_get_text(&mail_notify_entry);
	gkrellm_dup_string(&mail_notify, s);

	spin = GTK_SPIN_BUTTON(mail_fetch_spin_button);
	mail_fetch_timeout = gtk_spin_button_get_value_as_int(spin);

	spin = GTK_SPIN_BUTTON(mail_check_spin_button);
	mail_check_timeout = gtk_spin_button_get_value_as_int(spin);

	if (mailboxes_modified)
		{
		/* Destroy MBOX_INTERNAL type mailboxes from the mailbox_list, then
		|  recreate them from the clist.  Skip over the first FETCH mbox.
		*/
		for (list = mailbox_list->next; list;  )
			{
			mbox = (Mailbox *) list->data;
			if (mbox->type & MBOX_INTERNAL_FLAG)
				{
				GList *tmp;

/*				list = g_list_delete_link(list, list); */ /* glib 1.3 */
				tmp = list;
				list = g_list_remove_link(list, list);
				g_list_free(tmp);
				free_mailbox(mbox);
				}
			else
				list = list->next;
			}
		checking_mh_mail = FALSE;
		for (r = 0; r < GTK_CLIST(mbox_clist)->rows; ++r)
			{
			gtk_clist_get_text(GTK_CLIST(mbox_clist), r, 0, &s);
			mbox = add_mailbox(s);

			/* Restore old passwords from the shadowing password_list.
			*/
			if (mbox && mbox->password && *(mbox->password) == '*')
				{
				if (   (list = g_list_nth(password_list, r)) != NULL
					&& (s = (gchar *) list->data) != NULL
				   )
					{
					g_free(mbox->password);
					mbox->password = g_strdup(s);
					}
				else
					printf("Ooops: mail password list out of sync\n");
				}
			}
		force_mail_check = TRUE;
		}	
	}

static gchar	*mail_info_text[]	=
{
N_("<b>Mailboxes\n"),

#if defined (GKRELLM_THREADS)
N_("Mailbox files to monitor can be mbox, maildir, MH, POP3, or IMAP style\n"
"mailboxes.   The format for entering mailboxes to monitor in the "),
N_("<i>Mailboxes\n"),
N_(" tab is (in the same order as listed above):\n"),

N_("<i>\tfilename\n"),
N_("<i>\tdirectory\n"),
N_("<i>\tdirectory\n"),
N_("<i>\tmailserver   username   password\n"),
N_("<i>\tmailserver   mailbox   username   password\n"),

N_("\nSo, a simple example for monitoring your local mbox and a POP3 mailbox:\n"
),
N_("<i>\t/var/mail/bill\n"),
N_("<i>\tmail.gnu.org   bill2   PASSWORD\n"),
#else
N_("Mailbox files to monitor can be mbox or maildir style mailboxes.\n"),
#endif

N_("<b>\n\nSetup: Mail reading program\n"),
N_("If you enter a mail reading program (your mail user agent or MUA)\n"
"it can be launched by clicking on the mail monitor message count button.\n"
"For example you could enter:\n"),
N_("<i>\tcscmail\n"),
N_("or\n"),
N_("<i>\txterm -e mutt\n"),

N_("<b>\nSetup: Sound notify program\n"),
N_("If you enter a notify (sound) command, it will be run when new mail\n"
"is detected.  For example, you could enter:\n"),
N_("<i>\tcat /usr/local/sounds/newmail.wav > /dev/dsp\n"),
N_("or\n"),
N_("<i>\tesdplay /usr/local/sounds/newmail.wav\n\n"),

N_("<b>\nSetup: fetch/check Program\n"),
N_("If you want to download remote mail or check for remote mail without\n"
"using the builtin POP3 or IMAP mail checking which is set up in the\n"),
N_("<i>Mailboxes"),
N_(" tab, then do this by entering a mail fetch/check command.\n"
"For example, fetchmail can be run periodically to download mail\n"
"messages from a remote server to your local mailboxes where GKrellM\n"
"can count them.\n\n"

"Or, GKrellM can read the output and report the results from some mail\n"
"checking programs which count remote or local messages.  If you enter\n"
"a command like this that checks mail without downloading, then you\n"
"must tell GKrellM this is the expected behaviour by selecting in the\n"
"Options tab:\n"),
N_("<i>\tFetch/check program checks messages only\n"),
N_("For checking messages on a remote server, GKrellM can read the\n"
"output of the program "),
N_("<i>fetchmail -c"),
N_(" (you must append the -c).\n"),
N_("If you want to check for mail in local MH style mailboxes, then the\n"
"check program should be the nmh "),
N_("<i>flist -all.\n"),
N_("But, do not combine these methods for the same mailbox!  If you enter a\n"
"POP3 mailbox in the "),
N_("<i>Mailboxes"),
N_(" tab, then don't check it again with fetchmail.\n"),

N_("<b>\nMouse Button Actions:\n"),
N_("<b>\tLeft "),
N_("click the mail count button to launch the mail reading program.\n"),
N_("\t\tIf options permit, also stop animations and reset remote counts.\n"),
N_("<b>\tLeft "),
N_("click the envelope decal to force a mail check regardless of\n"
"\t\tany mute mode or inhibit mail checking option settings.\n"),
N_("<b>\tMiddle or Right "),
N_("click the mail panel to toggle a mute mode which inhibits\n"
	"\t\tthe sound notify program and optionally inhibits all mail\n"
	"\t\tchecking.\n"),
};

static gchar	*mbox_title[] =
	{N_("Mailbox files to monitor - see the Info page for entry formats")};

static void
create_mail_tab(GtkWidget *tab_vbox)
	{
	GtkWidget		*tabs;
	GtkWidget		*table;
	GtkWidget		*frame;
	GtkWidget		*vbox, *hbox;
	GtkWidget		*separator;
	GtkWidget		*label;
	GtkWidget		*button;
	GtkWidget		*scrolled;
	GtkWidget		*text;
	GList			*list;
	GSList			*group;
	gchar			*s, *hidden_passwd, *buf[2];
	gint			i, row;
	Mailbox			*mbox;

	mailboxes_modified = FALSE;
	selected_row = -1;
	free_glist_and_data(&password_list);
	mh_seq_ignore_button = NULL;

	tabs = gtk_notebook_new();
	gtk_notebook_set_tab_pos(GTK_NOTEBOOK(tabs), GTK_POS_TOP);
	gtk_box_pack_start(GTK_BOX(tab_vbox), tabs, TRUE, TRUE, 0);

/* --Setup tab */
	vbox = gkrellm_create_tab(tabs, _("Setup"));

	enable_mail_button = gtk_check_button_new_with_label(
				_("Enable Mailcheck"));
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(enable_mail_button),
				enable_mail);
	gtk_box_pack_start(GTK_BOX(vbox), enable_mail_button, TRUE, TRUE, 0);

	separator = gtk_hseparator_new();
	gtk_box_pack_start(GTK_BOX(vbox), separator, TRUE, TRUE, 0);

	table = gtk_table_new(5, 3, TRUE /*homogeneous*/);
	gtk_table_set_col_spacings(GTK_TABLE(table), 2);
	gtk_box_pack_start(GTK_BOX(vbox), table, FALSE, FALSE, 2);

	hbox = gtk_hbox_new(TRUE, 0);
	/* Attach left right top bottom */
	gtk_table_attach_defaults(GTK_TABLE(table), hbox, 0, 1, 0, 1);
	label = gtk_label_new(_("Mail reading program:"));
	gtk_misc_set_alignment(GTK_MISC(label), 1.0, 0.5);
	gtk_box_pack_start(GTK_BOX(hbox), label, TRUE, TRUE, 4);
	mail_user_agent_entry = gtk_entry_new_with_max_length(255);
	gtk_table_attach_defaults(GTK_TABLE(table), mail_user_agent_entry,
				1, 3, 0, 1);
	gtk_entry_set_text(GTK_ENTRY(mail_user_agent_entry),
				mail_user_agent.command);

	hbox = gtk_hbox_new(TRUE, 0);
	gtk_table_attach_defaults(GTK_TABLE(table), hbox, 0, 1, 1, 2);
	label = gtk_label_new(_("Notify (sound) program:"));
	gtk_misc_set_alignment(GTK_MISC(label), 1.0, 0.5);
	gtk_box_pack_start(GTK_BOX(hbox), label, TRUE, TRUE, 4);
	mail_notify_entry = gtk_entry_new_with_max_length(255);
	gtk_table_attach_defaults(GTK_TABLE(table), mail_notify_entry, 1, 3, 1, 2);
	gtk_entry_set_text(GTK_ENTRY(mail_notify_entry), mail_notify);

	hbox = gtk_hbox_new(TRUE, 0);
	gtk_table_attach_defaults(GTK_TABLE(table), hbox, 0, 1, 2, 3);
	label = gtk_label_new(_("Mail fetch/check program:"));
	gtk_misc_set_alignment(GTK_MISC(label), 1.0, 0.5);
	gtk_box_pack_start(GTK_BOX(hbox), label, TRUE, TRUE, 4);
	mail_fetch_entry = gtk_entry_new_with_max_length(255);
	gtk_table_attach_defaults(GTK_TABLE(table), mail_fetch_entry, 1, 3, 2, 3);
	gtk_entry_set_text(GTK_ENTRY(mail_fetch_entry),
				((Mailproc *)mail_fetch->private)->command);

	hbox = gtk_hbox_new(TRUE, 0);
	gtk_table_attach_defaults(GTK_TABLE(table), hbox, 0, 1, 3, 4);
	label = gtk_label_new(_("Check local mailboxes every"));
	gtk_misc_set_alignment(GTK_MISC(label), 1.0, 0.5);
	gtk_box_pack_start(GTK_BOX(hbox), label, TRUE, TRUE, 4);
	hbox = gtk_hbox_new(TRUE, 0);
	gtk_table_attach_defaults(GTK_TABLE(table), hbox, 1, 3, 3, 4);
	gkrellm_spin_button(hbox, &mail_check_spin_button,
			(gfloat) mail_check_timeout, 2.0, 100.0, 1.0, 5.0, 0, 60,
			NULL, NULL, FALSE, _("seconds"));

	hbox = gtk_hbox_new(TRUE, 0);
	gtk_table_attach_defaults(GTK_TABLE(table), hbox, 0, 1, 4, 5);
	label = gtk_label_new(_("Do remote checks every"));
	gtk_misc_set_alignment(GTK_MISC(label), 1.0, 0.5);
	gtk_box_pack_start(GTK_BOX(hbox), label, TRUE, TRUE, 4);
	hbox = gtk_hbox_new(TRUE, 0);
	gtk_table_attach_defaults(GTK_TABLE(table), hbox, 1, 3, 4, 5);
	gkrellm_spin_button(hbox, &mail_fetch_spin_button,
			(gfloat) mail_fetch_timeout, 1.0, 60.0, 1.0, 5.0, 0, 60,
			NULL, NULL, FALSE, _("minutes"));

	label = gtk_label_new(
_("Note: remote check interval applies to the \"Mail fetch/check program\"\n"
	"as well as to builtin remote POP3 and IMAP checking - see Info page"));
	gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_LEFT);
	gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 4);


/* --Mailboxes tab */
	vbox = gkrellm_create_tab(tabs, _("Mailboxes"));
	mbox_path_entry = gtk_entry_new_with_max_length(255);
	gtk_box_pack_start(GTK_BOX(vbox), mbox_path_entry, FALSE, FALSE, 2);
	gtk_entry_set_text(GTK_ENTRY(mbox_path_entry), "");
	gtk_signal_connect (GTK_OBJECT (mbox_path_entry), "activate",
			(GtkSignalFunc) cb_enter_mailbox, NULL);			
	hbox = gtk_hbox_new (FALSE, 0);
	gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);
	label = gtk_label_new("");
	gtk_box_pack_start(GTK_BOX(hbox), label, TRUE, TRUE, 5);
	button = gtk_button_new_with_label(_("Enter"));
	gtk_signal_connect(GTK_OBJECT(button), "clicked",
			(GtkSignalFunc) cb_enter_mailbox, NULL);
	gtk_box_pack_start(GTK_BOX(hbox), button, TRUE, TRUE, 5);
	button = gtk_button_new_with_label(_("Delete"));
	gtk_signal_connect(GTK_OBJECT(button), "clicked",
			(GtkSignalFunc) cb_delete_mailbox, NULL);
	gtk_box_pack_start(GTK_BOX(hbox), button, TRUE, TRUE, 5);

	separator = gtk_hseparator_new();
	gtk_box_pack_start(GTK_BOX(vbox), separator, FALSE, FALSE, 10);

	scrolled = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolled),
			GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	gtk_box_pack_start(GTK_BOX(vbox), scrolled, TRUE, TRUE, 0);
    *mbox_title = _(*mbox_title);
	mbox_clist = gtk_clist_new_with_titles(1, mbox_title);
	gtk_signal_connect (GTK_OBJECT(mbox_clist), "select_row",
		(GtkSignalFunc) cb_mailbox_selected, NULL);
	gtk_signal_connect (GTK_OBJECT(mbox_clist), "unselect_row",
		(GtkSignalFunc) cb_mailbox_unselected, NULL);
	gtk_container_add (GTK_CONTAINER (scrolled), mbox_clist);

	/* Skip the first internal mailbox.
	*/
	for (list = mailbox_list->next; list; list = list->next)
		{
		mbox = (Mailbox *) list->data;
		if (! (mbox->type & MBOX_INTERNAL_FLAG))
			continue;
		hidden_passwd = g_strdup(mbox->password);
		if (hidden_passwd)
			for (s = hidden_passwd; *s != '\0'; ++s)
				*s = '*';
		password_list = g_list_append(password_list, g_strdup(mbox->password));
		if (mbox->type == MBOX_POP3)
			buf[0] = g_strdup_printf("%s   %s   %s", mbox->server,
					mbox->username, hidden_passwd);
		else if (mbox->type == MBOX_IMAP)
			buf[0] = g_strdup_printf("%s   %s   %s   %s", mbox->server,
					mbox->folder, mbox->username, hidden_passwd);
		else
			buf[0] = g_strdup(mbox->path);

		buf[1] = NULL;
		row = gtk_clist_append(GTK_CLIST(mbox_clist), buf);
		g_free(buf[0]);
		if (hidden_passwd)
			g_free(hidden_passwd);
		}

/* --Options tab */
	vbox = gkrellm_create_tab(tabs, _("Options"));

	frame = gtk_frame_new (_("Animation Select"));
	gtk_box_pack_start(GTK_BOX(vbox), frame, TRUE, TRUE, 2);
	hbox = gtk_hbox_new (FALSE, 0);
	gtk_container_add(GTK_CONTAINER(frame), hbox);

	button = gtk_radio_button_new_with_label(NULL, _("None"));
	gtk_box_pack_start(GTK_BOX (hbox), button, TRUE, TRUE, 0);
	group = gtk_radio_button_group(GTK_RADIO_BUTTON (button));
	animation_select_button[0] = button;
	button = gtk_radio_button_new_with_label(group, _("Envelope"));
	gtk_box_pack_start(GTK_BOX (hbox), button, TRUE, TRUE, 0);
	group = gtk_radio_button_group(GTK_RADIO_BUTTON (button));
	animation_select_button[1] = button;
#ifdef BSD
	button = gtk_radio_button_new_with_label(group, _("Daemon"));
#else
	button = gtk_radio_button_new_with_label(group, _("Penguin"));
#endif
	gtk_box_pack_start(GTK_BOX(hbox), button, TRUE, TRUE, 0);
	group = gtk_radio_button_group(GTK_RADIO_BUTTON (button));
	animation_select_button[2] = button;
	button = gtk_radio_button_new_with_label(group, _("Both"));
	gtk_box_pack_start(GTK_BOX(hbox), button, TRUE, TRUE, 0);
	animation_select_button[3] = button;
	button = animation_select_button[animation_select_mode];
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button), TRUE);

	frame = gtk_frame_new (_("Message Counting"));
	gtk_box_pack_start(GTK_BOX(vbox), frame, TRUE, TRUE, 2);
	hbox = gtk_hbox_new (FALSE, 0);
	gtk_container_add(GTK_CONTAINER(frame), hbox);
	button = gtk_radio_button_new_with_label(NULL, _("new/total"));
	gtk_box_pack_start(GTK_BOX (hbox), button, TRUE, TRUE, 0);
	group = gtk_radio_button_group(GTK_RADIO_BUTTON (button));
	msg_count_button[0] = button;
	button = gtk_radio_button_new_with_label(group, _("new"));
	gtk_box_pack_start(GTK_BOX(hbox), button, TRUE, TRUE, 0);
	group = gtk_radio_button_group(GTK_RADIO_BUTTON (button));
	msg_count_button[1] = button;
	button = gtk_radio_button_new_with_label(group, _("do not count"));
	gtk_box_pack_start(GTK_BOX(hbox), button, TRUE, TRUE, 0);
	msg_count_button[2] = button;
	button = msg_count_button[msg_count_mode];
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button), TRUE);

    gkrellm_check_button(vbox, &enabel_cont_anim_button, cont_animation_mode,
		TRUE, 0,
		_("Run animation continuously as long as there is a new mail count"));

    gkrellm_check_button(vbox, &check_only_button, fetch_check_only_mode,
		TRUE, 0,
		_("Fetch/check program checks messages only (see Mail Info tab)"));

    gkrellm_check_button(vbox, &reset_remote_button, reset_remote_mode, TRUE,0,
		_("Reset remote message counts when message count button is pressed."));

    gkrellm_check_button(vbox, &super_mute_button, super_mute_mode, TRUE, 0,
   _("Mute mode inhibits all mail checking, not just notify (sound) program"));

    gkrellm_check_button(vbox, &mua_inhibit_button, mua_inhibit_mode, TRUE, 0,
		_("Inhibit all mail checking while the mail reader is running"));

    gkrellm_check_button(vbox, &show_tooltip_button, show_tooltip, TRUE, 0,
		_("List mailboxes containing new mail in a tooltip"));

	if (checking_mh_mail)
	    gkrellm_check_button(vbox, &mh_seq_ignore_button, mh_seq_ignore,
			TRUE, 0, _("Ignore .mh_sequences when checking MH mail."));
	
/* --Info tab */
	vbox = gkrellm_create_tab(tabs, _("Info"));
	scrolled = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolled),
			GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	gtk_box_pack_start(GTK_BOX(vbox), scrolled, TRUE, TRUE, 0);
	text = gtk_text_new(NULL, NULL);
	for (i = 0; i < sizeof(mail_info_text)/sizeof(gchar *); ++i)
		gkrellm_add_info_text_string(text, _(mail_info_text[i]));
	gtk_text_set_editable(GTK_TEXT(text), FALSE);
	gtk_container_add(GTK_CONTAINER(scrolled), text);
	}




static Monitor	monitor_mail =
	{
	N_("Mail"),				/* Name, for config tab.	*/
	MON_MAIL,			/* Id,  0 if a plugin		*/
	create_mail,		/* The create function		*/
	update_mail,		/* The update function		*/
	create_mail_tab,	/* The config tab create function	*/
	apply_mail_config,	/* Apply the config function		*/

	save_mail_config,	/* Save user conifg			*/
	load_mail_config,	/* Load user config			*/
	"mail",				/* config keyword			*/

	NULL,				/* Undef 2	*/
	NULL,				/* Undef 1	*/
	NULL,				/* Undef 0	*/

	0,					/* insert_before_id - place plugin before this mon */

	NULL,				/* Handle if a plugin, filled in by GKrellM		*/
	NULL				/* path if a plugin, filled in by GKrellM		*/
	};

Monitor *
init_mail_monitor(void)
	{
	monitor_mail.name = _(monitor_mail.name);
	enable_mail = TRUE;
	cont_animation_mode = FALSE;
	super_mute_mode = FALSE;
	mua_inhibit_mode = FALSE;

	mail_fetch = g_new0(Mailbox, 1);
	mail_fetch->private = g_new0(Mailproc, 1);
	mail_fetch->type = MBOX_FETCH;
	mail_fetch->check_func = run_fetch_program;
	mailbox_list = g_list_append(mailbox_list, mail_fetch);

	((Mailproc *)(mail_fetch->private))->command = g_strdup("");
	mail_user_agent.command = g_strdup("");
	mail_notify = g_strdup("");

	style_id = gkrellm_add_meter_style(&monitor_mail, MAIL_STYLE_NAME);
	force_mail_check = TRUE;	/* Do a check at startup */
	return &monitor_mail;
	}


  /* ====================== Exported Mail Functions =====================*/

gboolean
gkrellm_get_mail_mute_mode()
	{
	return mute_mode;
	}

  /* ====================================================================*/
  /* Functions to allow for external mailboxes, as in plugins which want
  |  to check for alternate mailbox types but have the mail counts reported
  |  by the builtin mailbox monitor.  A plugin outline would be:
  |
  |	typedef struct
  |		{
  |		gpointer mbox_ptr;
  |		... local plugin stuff ...
  |		} PluginMbox;
  |
  |	PluginMbox	pmbox;
  |
  |	void create_plugin_mbox(GtkWidget *vbox, gint first_create)
  |		{  //See plugin programmers reference for general structure
  |		...
  |		pmbox.mbox_ptr = gkrellm_add_external_mbox(pmbox_check, TRUE, &pmbox);
  |		gkrellm_set_external_mbox_tooltip(pmbox.mbox_ptr, "mbox_name_stuff");
  |		...
  |		}
  |
  |	gint pmbox_check(PluginMbox *pmb)
  |		{ //Collect info from the mail box
  |		gkrellm_set_external_mbox_counts(pmb->mbox_ptr, total_mail, new_mail);
  |		}
  */

  /* External mailbox counts won't show in the tooltip unless you call this.
  |  If a plugin wants only to use the sound/animation feature and not show
  |  up in a tooltip, then do not make this call.
  */
void
gkrellm_set_external_mbox_tooltip(gpointer mbox_ptr, gchar *string)
	{
	Mailbox *mbox = (Mailbox *) mbox_ptr;

	gkrellm_dup_string(&mbox->path, string);
	}

  /* Set total and new message counts for an external mailbox so the counts
  |  can appear in a tooltip (must also call above routine), and so the
  |  animation/sound can be triggered.  Since sound and animation is triggered
  |  synchronously with remote and local checks, a plugin should
  |  call this routine from within a check_func that is setup in the
  |  gkrellm_add_external_mbox() routine.
  */
void
gkrellm_set_external_mbox_counts(gpointer mbox_ptr, gint total, gint new)
	{
	Mailbox *mbox = (Mailbox *) mbox_ptr;

	mbox->mail_count = total;
	mbox->new_mail_count = new;
	}

  /* A plugin can have a check_func() called at the mail monitors local
  |  check interval (threaded is FALSE) or at the remote check interval
  |  (threaded is TRUE).  Additionally, if threaded is TRUE, the check_func
  |  will be called as a thread.  The data pointer is a pointer to a plugin
  |  defined structure which specifies a plugins unique mailbox.  This data
  |  pointer will be passed as the argument to check_func(data) when it
  |  is called at the update intervals.  gkrellm_add_external_mbox() returns
  |  the internal Mailbox * which should be treated simply as a gpointer
  |  in the plugin and it must be used as the first argument to the above
  |  gkrellm_set_external_tooltip() and gkrellm_set_external_counts().
  */
gpointer
gkrellm_add_external_mbox(gboolean (*check_func)(), gboolean threaded,
			gpointer data)
	{
	Mailbox	*mbox;

	mbox = g_new0(Mailbox, 1);
	mbox->type = threaded ? MBOX_EXTERNAL_THREADED : MBOX_EXTERNAL_INLINE;
	mbox->check_func = check_func;
	mbox->data = data;
	mailbox_list = g_list_append(mailbox_list, mbox);
	return (gpointer) mbox;
	}

void
gkrellm_destroy_external_mbox(gpointer mbox_ptr)
	{
	GList	*tmp;

	if ((tmp = g_list_find(mailbox_list, mbox_ptr)) == NULL)
		return;
	mailbox_list = g_list_remove_link(mailbox_list, tmp);
	g_list_free(tmp);
	free_mailbox(mbox_ptr);
	}

  /* ====================================================================*/
